# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from itertools import combinations

import pytest

from upstage_des.units.convert import CONVERSIONS, unit_convert


def test_convert_fail() -> None:
    with pytest.raises(ValueError):
        unit_convert(100, "parsec", "km")


def test_convert() -> None:
    result = unit_convert(100, "km", "mi")
    assert result == 0.62137119223 * 100

    result = unit_convert(3600, "s", "hr")
    assert result == 1.0

    result = unit_convert(10, "min", "s")
    assert result == 10.0 * 60.0


def test_convert_reverse() -> None:
    for unit_1, unit_2 in combinations(CONVERSIONS, 2):
        if unit_2 in CONVERSIONS[unit_1]:
            ans = unit_convert(1.0, unit_1, unit_2)
            reverse = unit_convert(ans, unit_2, unit_1)
            assert pytest.approx(reverse) == 1



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""Tests for nucleus and state sharing."""

from .flyer import Flyer, flyer_refuel_factory, mission_plan_net
from .mothership import Mothership, crew_factory, give_fuel_factory
from .mover import Mover, fly_end_factory

__all__ = [
    "Flyer",
    "flyer_refuel_factory",
    "mission_plan_net",
    "Mover",
    "fly_end_factory",
    "Mothership",
    "crew_factory",
    "give_fuel_factory",
]



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

import pytest

from upstage_des.actor import Actor
from upstage_des.api import EnvironmentContext, SimulationError, add_stage_variable
from upstage_des.data_types import GeodeticLocation
from upstage_des.geography import Spherical
from upstage_des.states import GeodeticLocationChangingState

# example lat lon alts
ATLANTA = [33.7490, -84.3880, 1050]
DENVER = [39.7392, -104.9903, 30_000]
SAN_FRAN = [37.7749, -122.4194, 0]


def test_create_geodetic() -> None:
    with EnvironmentContext():
        atl = GeodeticLocation(
            *ATLANTA,
        )
        san_fran = GeodeticLocation(
            *SAN_FRAN,
        )

        atl[0] == ATLANTA[0]
        with pytest.raises(IndexError):
            san_fran[3]

        with pytest.raises(ValueError):
            atl - 1

        with pytest.raises(ValueError):
            atl == 1

        assert atl != san_fran
        assert atl == atl
        assert san_fran == san_fran.copy()


def test_subtraction_geodetic() -> None:
    with EnvironmentContext():
        add_stage_variable("stage_model", Spherical)
        add_stage_variable("altitude_units", "ft")
        add_stage_variable("distance_units", "nmi")

        atl = GeodeticLocation(
            ATLANTA[0],
            ATLANTA[1],
            0.0,  # set feet to zero to test the distance against no altitude change
        )
        san_fran = GeodeticLocation(
            *SAN_FRAN,
        )
        dist = atl - san_fran
        assert abs(dist - 1857.8524277061492) <= 1e-6

        d = atl - atl.copy()
        assert d == 0

        atl = GeodeticLocation(*ATLANTA)
        den = GeodeticLocation(*DENVER)
        dist = atl.dist_with_altitude(den)
        assert abs(dist - 1052.67744200) <= 1e-6


def test_create_geodetic_changing() -> None:
    class StateTest(Actor):
        loc_state = GeodeticLocationChangingState(
            recording=True,
        )

    with EnvironmentContext():
        add_stage_variable("stage_model", Spherical)
        add_stage_variable("altitude_units", "ft")
        add_stage_variable("distance_units", "nmi")

        tester = StateTest(
            name="Geodetic Loc",
            loc_state=GeodeticLocation(*ATLANTA),
        )
        assert tester.loc_state == GeodeticLocation(*ATLANTA)
        assert tester.loc_state != GeodeticLocation(*DENVER)


def test_active_geodetic_changing() -> None:
    class StateTest(Actor):
        loc_state = GeodeticLocationChangingState(
            recording=True,
        )

    with EnvironmentContext():
        add_stage_variable("stage_model", Spherical)
        add_stage_variable("altitude_units", "ft")
        add_stage_variable("distance_units", "nmi")

        tester = StateTest(
            name="Geodetic Loc",
            loc_state=GeodeticLocation(*ATLANTA),
        )
        waypoints = [
            GeodeticLocation(*DENVER),
            GeodeticLocation(*SAN_FRAN),
        ]
        tester.activate_state(
            state="loc_state",
            task="dummy_task",  # type: ignore [arg-type]
            speed=1.0,
            waypoints=waypoints,
        )
        assert "loc_state" in tester._active_states
        assert tester._active_states["loc_state"] == tester.get_active_state_data("loc_state")

        # check the data created
        # ask for the state to update it
        tester.loc_state
        data = tester.get_active_state_data("loc_state")
        assert "path_data" in data
        assert len(data["path_data"]["times"]) == 2

        assert tester.loc_state == GeodeticLocation(*ATLANTA)
        tester.env.run(until=100)
        assert tester.loc_state != GeodeticLocation(*ATLANTA)
        assert tester.loc_state.alt > ATLANTA[2]

        # run time to reach Denver
        tester.env.run(until=1052.6666594454714)
        assert abs(tester.loc_state.lat - DENVER[0]) <= 1e-6
        assert abs(tester.loc_state.lon - DENVER[1]) <= 1e-6
        assert abs(tester.loc_state.alt - DENVER[2]) <= 1e-6

        # make sure it moves to the next waypoint
        tester.env.run(until=1200)
        tester.loc_state
        assert tester.loc_state.alt < DENVER[2]
        assert tester.loc_state.lon < DENVER[1]

        tester.env.run(until=1052.6666594454714 + 824.0883665214037)
        tester.loc_state
        assert abs(tester.loc_state.lat - SAN_FRAN[0]) <= 1e-6
        assert abs(tester.loc_state.lon - SAN_FRAN[1]) <= 1e-6
        assert abs(tester.loc_state.alt - SAN_FRAN[2]) <= 1e-6

        tester.env.run(until=tester.env.now + 10)
        with pytest.raises(SimulationError):
            tester.loc_state



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""Geographical types and protocols."""

from collections.abc import Callable
from dataclasses import dataclass
from typing import Protocol

POSITION = tuple[float, float, float]
POSITIONS = list[POSITION]
LAT_LON_ALT = POSITION
LAT_LON = tuple[float, float]


class GEO_POINT(Protocol):
    """Protocol for mimicking geographic data."""

    def latlon(self) -> LAT_LON:
        """Return a tuple of latitude and longitude in degrees."""


def _convert_geo(point: LAT_LON | GEO_POINT) -> LAT_LON:
    """Convert a geo point (or Lat/Lon) to Lat/Lon.

    Args:
        point (LAT_LON | GEO_POINT): The position on a globe

    Returns:
        LAT_LON: Position as correct type.
    """
    if isinstance(point, tuple):
        return point
    return point.latlon()


class EarthProtocol(Protocol):
    """Protocol for defining an earth model interface."""

    def distance(
        self,
        loc1: LAT_LON | GEO_POINT,
        loc2: LAT_LON | GEO_POINT,
        units: str,
    ) -> float:
        """Get the distance between two lat/lon (degrees) points."""

    def bearing(
        self,
        loc1: LAT_LON | GEO_POINT,
        loc2: LAT_LON | GEO_POINT,
    ) -> float:
        """Get the distance between two lat/lon (degrees) points."""

    def distance_and_bearing(
        self,
        loc1: LAT_LON | GEO_POINT,
        loc2: LAT_LON | GEO_POINT,
        units: str,
    ) -> tuple[float, float]:
        """Get the distance between two lat/lon (degrees) points."""

    def point_from_bearing_dist(
        self,
        point: LAT_LON | GEO_POINT,
        bearing: float,
        distance: float,
        distance_units: str = "nmi",
    ) -> tuple[float, float]:
        """Get a lat/lon in degrees from a point, bearing, and distance."""

    def lla2ecef(
        self,
        locs: list[LAT_LON_ALT],
    ) -> list[tuple[float, float, float]]:
        """Get ECEF coordinates from lat lon alt."""

    def ecef2lla(
        self,
        locs: list[LAT_LON_ALT],
    ) -> list[tuple[float, float, float]]:
        """Get ECEF coordinates from lat lon alt."""

    def geo_linspace(
        self,
        start: LAT_LON | GEO_POINT,
        end: LAT_LON | GEO_POINT,
        num_segments: int,
    ) -> list[LAT_LON]:
        """Get evenly spaced coordinates between lat/lon pairs."""

    def geo_circle(
        self,
        center: LAT_LON | GEO_POINT,
        radius: float,
        radius_units: str,
        num_points: int,
    ) -> list[LAT_LON]:
        """Create a circle on a globe."""


@dataclass
class CrossingCondition:
    """Data about an intersection."""

    kind: str
    begin: LAT_LON_ALT
    end: LAT_LON_ALT | None = None


INTERSECTION_LOCATION_CALLABLE = Callable[
    [
        LAT_LON_ALT,
        LAT_LON_ALT,
        LAT_LON_ALT,
        float,
        str,
        EarthProtocol,
        float | None,
        list[int] | None,
    ],
    list[CrossingCondition],
]



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from collections.abc import Callable
from typing import Any

import pytest
from simpy import Environment, Store

from upstage_des.actor import Actor
from upstage_des.base import EnvironmentContext
from upstage_des.events import FilterGet
from upstage_des.resources.monitoring import (
    SelfMonitoringContainer,
    SelfMonitoringFilterStore,
    SelfMonitoringReserveContainer,
    SelfMonitoringSortedFilterStore,
    SelfMonitoringStore,
)
from upstage_des.resources.reserve import ReserveContainer
from upstage_des.resources.sorted import SortedFilterGet, SortedFilterStore
from upstage_des.type_help import SIMPY_GEN

MAX_RUN_TIME = 10.0


def getter(
    env: Environment, store: Store, wait: float = 1.0, cback: Callable | None = None, **kwargs: Any
) -> SIMPY_GEN:
    yield env.timeout(wait)
    get = store.get(**kwargs)
    if cback is not None:
        get.callbacks.append(cback)
    item = yield get
    return item


def sorted_filter_getter(
    env: Environment,
    store: SortedFilterStore,
    wait: float,
    filter: Callable[[Any], bool],
    sorter: Callable[[Any], Any] | None = None,
    reverse: bool = False,
) -> SIMPY_GEN:
    yield env.timeout(wait)
    evt = SortedFilterGet(
        store,
        filter,
        sorter,
        reverse,
    )
    item = yield evt.as_event()
    return item


def putter(
    env: Environment,
    store: Store,
    item: Any,
    wait: float = 0.0,
    cback: Callable | None = None,
    **kwargs: Any,
) -> SIMPY_GEN:
    yield env.timeout(wait)
    put = store.put(item, **kwargs)
    if cback is not None:
        put.callbacks.append(cback)
    yield put


def test_notifying_store() -> None:
    notifications: list[Any] = []

    def callback(*args: Any, **kwargs: Any) -> None:
        assert len(kwargs) == 0
        notifications.append(args)

    with EnvironmentContext() as env:
        store = Store(env=env)

        def sim() -> SIMPY_GEN:
            item = "an item"
            yield env.process(putter(env, store, item, cback=callback))
            retrieved_item = yield env.process(getter(env, store, cback=callback))
            assert item == retrieved_item

        env.process(sim())

        env.run(until=MAX_RUN_TIME)

        assert len(notifications) == 2


def test_sorted_filter_store() -> None:
    with EnvironmentContext() as env:
        store = SortedFilterStore(env=env)

        def get_proc() -> SIMPY_GEN:
            item = yield env.process(
                getter(
                    env,
                    store=store,
                    filter=lambda x: isinstance(x, int | float),
                    sorter=lambda x: (-x,),
                    wait=0.0,
                )
            )
            return item

        def sim() -> SIMPY_GEN:
            env.process(putter(env, store, 10, wait=0.0))
            env.process(putter(env, store, 1, wait=0.0))

            item = yield env.process(get_proc())

            assert item == 10

        env.process(sim())

        env.run(until=MAX_RUN_TIME)


def test_sorted_filter_store_upstage_get() -> None:
    with EnvironmentContext() as env:
        store = SortedFilterStore(env=env)

        def get_proc() -> SIMPY_GEN:
            item = yield env.process(
                sorted_filter_getter(
                    env,
                    store=store,
                    filter=lambda x: isinstance(x, int | float),
                    sorter=lambda x: x,
                    reverse=True,
                    wait=0.0,
                )
            )
            return item

        def sim() -> SIMPY_GEN:
            env.process(putter(env, store, 10, wait=0.0))
            env.process(putter(env, store, 1, wait=0.0))

            item = yield env.process(get_proc())

            assert item == 10

        env.process(sim())

        env.run(until=MAX_RUN_TIME)


def test_reserve_store() -> None:
    with EnvironmentContext() as env:
        store = ReserveContainer(
            env=env,
            capacity=10,
            init=10,
        )
        assert store.available == 10

        class R(Actor): ...

        requestor = R(name="wanter")

        req = store.reserve(requestor, 8, expiration=12)
        assert req is not False
        req1 = store.reserve(R(name="other"), 8)
        assert not req1

        assert store.available == 2
        assert len(store._queued) == 1
        env.run(until=1)
        store.cancel_request(requestor)
        assert len(store._queued) == 0
        assert store.available == 10

        req = store.reserve(requestor, 8, expiration=8)
        assert req is not False
        env.run(until=13)
        assert len(store.queued) == 0
        assert store.available == 10
        with pytest.raises(ValueError):
            store.take(requestor)

        req = store.reserve(requestor, 8, expiration=8)
        env.run(until=14)
        assert store.take(requestor) == 8
        env.run(until=30)
        assert store.available == 2

        with pytest.raises(ValueError):
            store.put(100, capacity_increase=False)

        store.put(3)
        assert store.available == 5


def test_self_monitoring_filter_store() -> None:
    with EnvironmentContext() as env:
        store = SelfMonitoringFilterStore(env=env)

        def filter(item: str) -> bool:
            return "Another" in item

        def proc() -> SIMPY_GEN:
            yield store.put("The Item")
            yield store.put("Another Item")
            evt = FilterGet(
                store,
                filter,
            )
            item = yield evt.as_event()
            return item

        def sim() -> SIMPY_GEN:
            item = yield env.process(proc())
            assert item == "Another Item"

        env.process(sim())

        env.run(until=MAX_RUN_TIME)


def test_self_monitoring_sorted_filter_store() -> None:
    with EnvironmentContext() as env:
        store = SelfMonitoringSortedFilterStore(env=env)

        def get_proc() -> SIMPY_GEN:
            item = yield env.process(
                getter(
                    env,
                    store=store,
                    filter=lambda x: isinstance(x, int | float),
                    sorter=lambda x: (-x,),
                    wait=0.0,
                ),
            )
            return item

        def sim() -> SIMPY_GEN:
            env.process(putter(env, store, 10, wait=0.0))
            env.process(putter(env, store, 1, wait=0.0))

            item = yield env.process(get_proc())

            assert item == 10

        env.process(sim())

        env.run(until=MAX_RUN_TIME)


def test_self_monitoring_store() -> None:
    with EnvironmentContext() as env:
        SelfMonitoringStore(env)
        env.run(until=MAX_RUN_TIME)


def test_self_monitoring_reserve_store() -> None:
    with EnvironmentContext() as env:
        SelfMonitoringReserveContainer(env)
        env.run(until=MAX_RUN_TIME)


def test_self_monitoring_container() -> None:
    with EnvironmentContext() as env:
        con = SelfMonitoringContainer(env, capacity=10)

        def sim() -> SIMPY_GEN:
            evt = con.put(5)
            yield evt
            evt = con.get(3)
            yield evt

        env.process(sim())
        env.run()
        assert len(con._quantities) == 3
        assert [0, 5, 2] == [x[1] for x in con._quantities]



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""This module contains math utility functions to avoid numpy."""

from math import sqrt

VECTOR = list[float] | tuple[float, ...]


def _vector_subtract(vector_a: VECTOR, vector_b: VECTOR) -> list[float]:
    """Subtract equal-sized vectors.

    Args:
        vector_a (list[float]): Left vector
        vector_b (list[float]): Right vector

    Returns:
        list[float]: Subtracted vector
    """
    if not (len(vector_a) == len(vector_b)):
        raise ValueError("Vectors are not the same size")

    ret = [a - b for a, b in zip(vector_a, vector_b)]
    return ret


def _vector_add(vector_a: VECTOR, vector_b: VECTOR) -> list[float]:
    """Add equal-sized vectors.

    Args:
        vector_a (list[float]): Left vector
        vector_b (list[float]): Right vector

    Returns:
        list[float]: Added vector
    """
    if not (len(vector_a) == len(vector_b)):
        raise ValueError("Vectors are not the same size")

    ret = [a + b for a, b in zip(vector_a, vector_b)]
    return ret


def _vector_dot(vector_a: VECTOR, vector_b: VECTOR) -> float:
    """Inner product of two vectors.

    Args:
        vector_a (VECTOR): Left vector
        vector_b (VECTOR): Right vector

    Returns:
        float: inner product
    """
    return sum(a * b for a, b in zip(vector_a, vector_b))


def _vector_norm(arr: VECTOR) -> float:
    """Norm of a vector.

    Args:
        arr (VECTOR): vector

    Returns:
        float: norm
    """
    s = sum(a**2 for a in arr)
    return sqrt(s)


def _roots(a: float, b: float, c: float) -> list[float]:
    """Calculate the roots of a quadratic.

    The form is ax^2 + bx + c = 0.

    Args:
        a (float): Coefficient on the square term
        b (float): Coefficient on the base term
        c (float): Constant

    Returns:
        list[float]: The two roots, empty if not real.
    """
    discriminant = b**2 - 4 * a * c
    if discriminant < 0:
        return []
    root1 = (-b + sqrt(discriminant)) / (2 * a)
    root2 = (-b - sqrt(discriminant)) / (2 * a)
    return [root1, root2]


def _col_mat_mul(col: VECTOR, mat: list[list[float]]) -> list[float]:
    """Do a matrix multiplication of a column against a matrix.

    Does col @ mat

    Args:
        col (VECTOR): The left vector
        mat (list[list[float]]): The matrix

    Returns:
        list[float]: The multiplied result
    """
    if len(col) != len(mat):
        raise ValueError("Number of values in col must equal number of rows in mat")

    result = [sum(x * y for x, y in zip(col, c)) for c in zip(*mat)]

    return result



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""WGS84 Earth model."""

from math import atan, atan2, cos, degrees, radians, sin, sqrt, tan

from upstage_des.units import unit_convert

from .conversions import WGS84_A, WGS84_B, WGS84_F, WGS84Conversions
from .geo_types import GEO_POINT, LAT_LON, POSITIONS, _convert_geo


class WGS84(WGS84Conversions):
    """Geographical math using elliptical coordinates for Earth.

    Based on Vincenty's methods and WGS84 parameters.
    """

    @classmethod
    def distance(
        cls,
        loc1: LAT_LON | GEO_POINT,
        loc2: LAT_LON | GEO_POINT,
        units: str = "nmi",
        tol: float = 1e-12,
        max_iter: int = 200,
    ) -> float:
        """Find distance between two points.

        Return the great circle distance from the loc1 to another loc2.

        Args:
            loc1 (LAT_LON): Starting point (degrees)
            loc2 (LAT_LON): Ending point (degrees)
            units (str, optional): Units requested for distance. Defaults to 'nmi'.
            tol (float, optional): Calculation convergence tolerance. Defaults to 1e-12
            max_iter (int, optional): Max iterations. Defaults to 200

        Returns:
            float: Distances in the defined units
        """
        dist, _ = cls.distance_and_bearing(loc1, loc2, units, tol, max_iter)
        return dist

    @classmethod
    def bearing(
        cls,
        loc1: LAT_LON | GEO_POINT,
        loc2: LAT_LON | GEO_POINT,
        tol: float = 1e-12,
        max_iter: int = 200,
    ) -> float:
        """Calculate the forward bearing (in degrees) from loc1 to loc2.

        Args:
            loc1 (LAT_LON): Starting point (degrees)
            loc2 (LAT_LON): Ending point (degrees)
            units (str, optional): Units requested for distance. Defaults to 'nmi'.
            tol (float, optional): Calculation convergence tolerance. Defaults to 1e-12
            max_iter (int, optional): Max iterations. Defaults to 200

        Returns:
            float: Bearing
        """
        _, angle = cls.distance_and_bearing(loc1, loc2, tol=tol, max_iter=max_iter)
        return angle

    @classmethod
    def distance_and_bearing(
        cls,
        loc1: LAT_LON | GEO_POINT,
        loc2: LAT_LON | GEO_POINT,
        units: str = "nmi",
        tol: float = 1e-12,
        max_iter: int = 200,
    ) -> tuple[float, float]:
        """Find great circle distance and forward bearing between two points.

        Args:
            loc1 (LAT_LON): Starting point
            loc2 (LAT_LON): Ending point
            units (str, optional): Distance units requested. Defaults to 'nmi'.
            tol (float, optional): Calculation convergence tolerance. Defaults to 1e-12
            max_iter (int, optional): Max iterations. Defaults to 200

        Returns:
            tuple[float, float]: Distance and Bearing
        """
        loc1 = _convert_geo(loc1)
        loc2 = _convert_geo(loc2)
        if loc1[0] == loc2[0] and loc1[1] == loc2[1]:
            return 0.0, 0.0
        # reduced latitudes
        u1 = atan((1 - WGS84_F) * tan(radians(loc1[0])))
        u2 = atan((1 - WGS84_F) * tan(radians(loc2[0])))
        delta_lon = radians(loc2[1] - loc1[1])
        lambda_lon = delta_lon

        sin_u1 = sin(u1)
        cos_u1 = cos(u1)
        sin_u2 = sin(u2)
        cos_u2 = cos(u2)

        for _ in range(max_iter):
            sin_lambda = sin(lambda_lon)
            cos_lambda = cos(lambda_lon)
            sin_sigma = sqrt(
                (cos_u2 * sin_lambda) ** 2 + (cos_u1 * sin_u2 - sin_u1 * cos_u2 * cos_lambda) ** 2
            )

            cos_sigma = sin_u1 * sin_u2 + cos_u1 * cos_u2 * cos_lambda
            sigma = atan2(sin_sigma, cos_sigma)
            sin_alpha = cos_u1 * cos_u2 * sin_lambda / sin_sigma
            cos_sq_alpha = 1 - sin_alpha**2
            try:
                cos2_sigma_m = cos_sigma - 2 * sin_u1 * sin_u2 / cos_sq_alpha
            except ZeroDivisionError:
                cos2_sigma_m = 0
            c = WGS84_F / 16 * cos_sq_alpha * (4 + WGS84_F * (4 - 3 * cos_sq_alpha))
            _lambda_prev = lambda_lon
            lambda_lon = delta_lon + (1 - c) * WGS84_F * sin_alpha * (
                sigma + c * sin_sigma * (cos2_sigma_m + c * cos_sigma * (-1 + 2 * cos2_sigma_m**2))
            )
            if abs(lambda_lon - _lambda_prev) < tol:
                break  # successful convergence
        else:
            raise ValueError("Could not converge distance/bearing calculation.")

        u_sq = cos_sq_alpha * (WGS84_A**2 - WGS84_B**2) / (WGS84_B**2)
        a = 1 + u_sq / 16384 * (4096 + u_sq * (-768 + u_sq * (320 - 175 * u_sq)))
        b = u_sq / 1024 * (256 + u_sq * (-128 + u_sq * (74 - 47 * u_sq)))
        delta_sigma = (
            b
            * sin_sigma
            * (
                cos2_sigma_m
                + b
                / 4
                * (
                    cos_sigma * (-1 + 2 * cos2_sigma_m**2)
                    - b / 6 * cos2_sigma_m * (-3 + 4 * sin_sigma**2) * (-3 + 4 * cos2_sigma_m**2)
                )
            )
        )

        dist_m = round(WGS84_B * a * (sigma - delta_sigma), 6)

        azimuth = atan2(
            cos_u2 * sin_lambda,
            cos_u1 * sin_u2 - sin_u1 * cos_u2 * cos_lambda,
        )
        return unit_convert(dist_m, "m", units), degrees(azimuth) % 360

    @classmethod
    def point_from_bearing_dist(
        cls,
        point: LAT_LON | GEO_POINT,
        bearing: float,
        distance: float,
        distance_units: str = "nmi",
        tol: float = 1e-12,
        max_iter: int = 200,
    ) -> LAT_LON:
        """Get a Location from a starting point, a bearing, and a distance.

        Args:
            point (LAT_LON): Starting point (degrees)
            bearing (float): Bearing to travel along (degrees)
            distance (float): Distance to travel
            distance_units (str, optional): Units of the distance. Defaults to 'nmi'.
            tol (float, optional): Calculation convergence tolerance. Defaults to 1e-12
            max_iter (int, optional): Max iterations. Defaults to 200

        Returns:
            LAT_LON: The point in degrees
        """
        point = _convert_geo(point)
        s = unit_convert(distance, distance_units, "km") * 1000
        phi_1 = radians(point[0])
        lambda_1 = radians(point[1])
        a, b, f = WGS84_A, WGS84_B, WGS84_F
        alpha_1 = radians(bearing)

        sin_alpha_1 = sin(alpha_1)
        cos_alpha_1 = cos(alpha_1)

        tan_u1 = (1 - f) * tan(phi_1)
        cos_u1 = 1 / sqrt(1 + tan_u1**2)
        sin_u1 = tan_u1 * cos_u1

        sigma_1 = atan2(tan_u1, cos_alpha_1)
        sin_alpha = cos_u1 * sin_alpha_1
        cos_sq_alpha = 1 - sin_alpha**2
        u_sq = cos_sq_alpha * (a**2 - b**2) / (b**2)
        A = 1 + u_sq / 16384 * (4096 + u_sq * (-768 + u_sq * (320 - 175 * u_sq)))
        B = u_sq / 1024 * (256 + u_sq * (-128 + u_sq * (74 - 47 * u_sq)))

        sigma = s / (b * A)
        for _ in range(max_iter):
            cos2_sigma_m = cos(2 * sigma_1 + sigma)
            sin_sigma = sin(sigma)
            cos_sigma = cos(sigma)
            delta_sigma = (
                B
                * sin_sigma
                * (
                    cos2_sigma_m
                    + B
                    / 4
                    * (
                        cos_sigma * (-1 + 2 * cos2_sigma_m**2)
                        - B
                        / 6
                        * cos2_sigma_m
                        * (-3 + 4 * sin_sigma**2)
                        * (-3 + 4 * cos2_sigma_m**2)
                    )
                )
            )
            sigma_prime = sigma
            sigma = s / (b * A) + delta_sigma
            if abs(sigma_prime - sigma) < tol:
                break
        else:
            raise ValueError("Failed to converge on point from bearing and distance.")

        x = sin_u1 * sin_sigma - cos_u1 * cos_sigma * cos_alpha_1
        phi_2 = atan2(
            sin_u1 * cos_sigma + cos_u1 * sin_sigma * cos_alpha_1,
            (1 - f) * sqrt(sin_alpha**2 + x**2),
        )
        lam = atan2(
            sin_sigma * sin_alpha_1,
            cos_u1 * cos_sigma - sin_u1 * sin_sigma * cos_alpha_1,
        )
        C = f / 16 * cos_sq_alpha * (4 + f * (4 - 3 * cos_sq_alpha))
        L = lam - (1 - C) * f * sin_alpha * (
            sigma + C * sin_alpha * (cos2_sigma_m + C * cos_sigma * (-1 + 2 * cos2_sigma_m**2))
        )
        lambda_2 = lambda_1 + L

        lat, lon = degrees(phi_2), degrees(lambda_2)
        return lat, lon

    @classmethod
    def geo_linspace(
        cls,
        start: LAT_LON | GEO_POINT,
        end: LAT_LON | GEO_POINT,
        num_segments: int = 10,
    ) -> list[LAT_LON]:
        """Make a discrete great circle path between two points.

        Altitudes are not considered since they do not affect angular coordinates.

        The number of points is the number

        Args:
            start (LAT_LON): Start point
            end (LAT_LON): End point
            num_segments (int, optional): Number of segments on the path. Defaults to 10.

        Returns:
           list[LAT_LON]: Lattiude and Longitude (degrees)
        """
        start = _convert_geo(start)
        end = _convert_geo(end)
        if num_segments < 2:
            raise ValueError("Not enough points for interpolation")

        # total distance in km
        dist, bearing = cls.distance_and_bearing(start, end, units="km")

        lats = []
        lons = []
        for fraction in range(0, num_segments + 1):
            f = fraction / num_segments
            if fraction == 0:
                loc = start
            elif fraction == num_segments:
                loc = end
            else:
                loc = cls.point_from_bearing_dist(
                    start,
                    bearing,
                    dist * f,
                    distance_units="km",
                )
            lats.append(loc[0])
            lons.append(loc[1])

        return [(la, lo) for la, lo in zip(lats, lons)]

    @classmethod
    def geo_circle(
        cls,
        center: LAT_LON | GEO_POINT,
        radius: float,
        radius_units: str = "nmi",
        num_points: int = 10,
    ) -> list[LAT_LON]:
        """Make an approximate circle by sweeping bearing from a central location.

        Altitudes are not considered since they do not affect angular coordinates.

        Args:
            center (LAT_LON): Center point (degrees)
            radius (float): Radius of the circle
            radius_units (str, optional): Units of the radius. Defaults to 'nmi'.
            num_points (int, optional): Number of points on the circle. Defaults to 10.

        Returns:
            list[LAT_LON]: Latitude and Longitude (degrees) of the circle
        """
        center = _convert_geo(center)
        lats = []
        lons = []
        for fraction in range(0, num_points + 1):
            f = fraction / num_points
            bearing = f * 360
            loc = cls.point_from_bearing_dist(
                center,
                bearing,
                radius,
                distance_units=radius_units,
            )
            lats.append(loc[0])
            lons.append(loc[1])

        return [(la, lo) for la, lo in zip(lats, lons)]

    @classmethod
    def ecef_linspace(
        cls,
        start: LAT_LON | GEO_POINT,
        finish: LAT_LON | GEO_POINT,
        start_alt: float,
        finish_alt: float,
        segments: int,
    ) -> POSITIONS:
        """Get an array of spaced points along a greate circle path in ECEF.

        Args:
            start (LAT_LON): Start point (lat/lon degrees)
            finish (LAT_LON): End point (lat/lon degrees)
            start_alt (float): Starting altitude (meters)
            finish_alt (float): Ending altitude (meters)
            segments (int): Number of segments to create

        Returns:
            POSITIONS: ECEF coordinates
        """
        start = _convert_geo(start)
        finish = _convert_geo(finish)
        lats, lons = cls.geo_linspace(start, finish, num_segments=segments)
        delta_alt = finish_alt - start_alt
        alts = [start_alt + delta_alt * i / segments for i in range(segments + 1)]
        lla = [(lat, lon, alt) for lat, lon, alt in zip(lats, lons, alts)]
        return cls.lla2ecef(lla)

    @classmethod
    def ecef_and_geo_linspace(
        cls,
        start: LAT_LON | GEO_POINT,
        finish: LAT_LON | GEO_POINT,
        start_alt: float,
        finish_alt: float,
        segments: int,
    ) -> tuple[POSITIONS, POSITIONS]:
        """Get an array of spaced points along a greate circle path in ECEF and LLA.

        Args:
            start (LAT_LON): Start point (lat/lon degrees)
            finish (LAT_LON): End point (lat/lon degrees)
            start_alt (float): Starting altitude (meters)
            finish_alt (float): Ending altitude (meters)
            segments (int): Number of segments to create

        Returns:
            POSITIONS: ECEF coordinates
            list[tuple[float, float, float]]: Lat/Lon/Alt (degrees)
        """
        start = _convert_geo(start)
        finish = _convert_geo(finish)
        latlons = cls.geo_linspace(start, finish, num_segments=segments)
        delta_alt = finish_alt - start_alt
        alts = [start_alt + delta_alt * i / segments for i in range(segments + 1)]
        lla: POSITIONS = [(latlon[0], latlon[1], alt) for latlon, alt in zip(latlons, alts)]
        return cls.lla2ecef(lla), lla



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

import upstage_des.api as UP


class Example(UP.Actor):
    a_value = UP.State[float](default=1.0)


class Example2(Example):
    b_value = UP.State[float](default=2.0)


class EnvHolder(UP.UpstageBase): ...


def test_actor_tracking() -> None:
    with UP.EnvironmentContext():
        env = EnvHolder()
        actor = Example(name="Example_1")

        actor2 = Example2(
            name="Example_2",
            b_value=3.0,
        )

        actors = env.get_actors()
        assert len(actors) == 2, "Wrong number of actors stored"
        assert actor in actors
        assert actor2 in actors

        actor_entities = env.get_entity_group("Example")
        assert actor in actor_entities
        assert len(actor_entities) == 2

        actor_entities = env.get_entity_group("Example2")
        assert actor2 in actor_entities
        assert len(actor_entities) == 1


def test_env_reset() -> None:
    with UP.EnvironmentContext():
        env = EnvHolder()
        actor = Example(name="Example_1")

        actors = env.get_actors()
        assert len(actors) == 1, "Wrong number of actors stored"
        assert actor in actors

    with UP.EnvironmentContext():
        env = EnvHolder()
        actors = env.get_actors()
        assert actor not in actors

        actors = env.get_actors()
        assert actor not in actors


def test_actor_multi_tracking() -> None:
    class Person(UP.Actor, entity_groups=("A Person", "VIP")):
        pass

    with UP.EnvironmentContext():
        env = EnvHolder()
        p = Person(name="some person")
        people = env.get_entity_group("A Person")
        assert len(people) == 1
        vips = env.get_entity_group("VIP")
        assert len(vips) == 1
        actors = env.get_actors()
        assert len(actors) == 1
        assert p in actors
        assert len(env.get_all_entity_groups()) == 3

        class Doctor(UP.Actor, entity_groups=("Hospital Worker")):
            pass

        d = Doctor(name="Doc Oc")
        assert len(env.get_all_entity_groups()) == 5
        assert d not in env.get_entity_group("Person")


class Sensor(UP.NamedUpstageEntity):
    def __init__(self, name: str, radius: float) -> None:
        self.name = name
        self.radius = radius


class RadarSensor(Sensor, entity_groups="RADAR"):
    pass


class LightSensor(Sensor):
    pass


class NextRadar(RadarSensor):
    pass


def test_entity_tracking() -> None:
    with UP.EnvironmentContext():
        env = EnvHolder()
        simple_sensor = Sensor(name="Sense1", radius=3)
        radar_sensor = RadarSensor(name="A Radar", radius=10)
        light_sensor = LightSensor(name="A Light", radius=4.5)

        sensors = env.get_entity_group("Sensor")
        radars = env.get_entity_group("RADAR")
        other_radars = env.get_entity_group("RadarSensor")
        lights = env.get_entity_group("LightSensor")

        assert len(sensors) == 3
        assert simple_sensor in sensors
        assert len(radars) == 1
        assert radar_sensor in radars
        assert len(lights) == 1
        assert light_sensor in lights
        assert len(other_radars) == 1

        assert len(env.get_all_entity_groups()) == 4

        # Check that we do inherit subclass and the defined ones.
        nxt = NextRadar(name="More testing", radius=6.45)
        assert nxt in env.get_entity_group("RADAR")
        assert nxt in env.get_entity_group("Sensor")
        assert nxt in env.get_entity_group("RadarSensor")


def test_multi_tracking() -> None:
    with UP.EnvironmentContext():
        env = EnvHolder()

        class LightSensor(RadarSensor, entity_groups=("MySensor", "Light")):
            pass

        LightSensor(name="A Light", radius=4.5)
        lights = env.get_entity_group("Light")
        sensors = env.get_entity_group("MySensor")
        assert len(lights) == 1
        assert len(sensors) == 1


def test_env_reset_tracking() -> None:
    with UP.EnvironmentContext():
        env = EnvHolder()
        simple_sensor = Sensor(name="Sense1", radius=3)
        sensors = env.get_entity_group("Sensor")
        assert len(sensors) == 1
        assert simple_sensor in sensors

    with UP.EnvironmentContext():
        env = EnvHolder()
        sensors = env.get_entity_group("Sensor")
        assert simple_sensor not in sensors

        sensors = env.get_entity_group("Sensor")
        assert simple_sensor not in sensors
        assert len(env.get_all_entity_groups()) == 0



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from upstage_des import api


def test_api() -> None:
    api_items = dir(api)

    items_to_test = (
        "UpstageError",
        "SimulationError",
        "RulesError",
        "Actor",
        "PLANNING_FACTOR_OBJECT",
        "UpstageBase",
        "NamedUpstageEntity",
        "EnvironmentContext",
        "add_stage_variable",
        "get_stage_variable",
        "get_stage",
        "All",
        "Any",
        "Event",
        "Get",
        "FilterGet",
        "SortedFilterGet",
        "Put",
        "ResourceHold",
        "Wait",
        "ContainerEmptyError",
        "ContainerError",
        "ContainerFullError",
        "ContinuousContainer",
        "SelfMonitoringContainer",
        "SelfMonitoringContinuousContainer",
        "SelfMonitoringFilterStore",
        "SelfMonitoringSortedFilterStore",
        "SelfMonitoringReserveContainer",
        "SelfMonitoringStore",
        "ReserveContainer",
        "SortedFilterStore",
        "Location",
        "CartesianLocation",
        "GeodeticLocation",
        "CartesianLocationData",
        "GeodeticLocationData",
        "LinearChangingState",
        "CartesianLocationChangingState",
        "State",
        "GeodeticLocationChangingState",
        "DetectabilityState",
        "ResourceState",
        "DecisionTask",
        "Task",
        "process",
        "InterruptStates",
        "TerminalTask",
        "TaskNetwork",
        "TaskNetworkFactory",
        "TaskLinks",
        "CommsManager",
        "Message",
        "MessageContent",
        "MotionAndDetectionError",
        "SensorMotionManager",
        "SteppedMotionManager",
        "TaskNetworkNucleus",
        "NucleusInterrupt",
        "SharedLinearChangingState",
        "CommunicationStore",
        "unit_convert",
    )

    for item in items_to_test:
        assert item in api_items

    for item in api_items:
        if not item.startswith("_"):
            assert item in items_to_test



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""Base classes and exceptions for UPSTAGE."""

from collections import defaultdict
from collections.abc import Iterable
from contextvars import ContextVar, Token
from dataclasses import dataclass, field
from functools import wraps
from math import floor
from random import Random
from time import gmtime, strftime
from typing import TYPE_CHECKING, Any, Protocol, Union
from warnings import warn

from simpy import Environment as SimpyEnv

from upstage_des.geography import INTERSECTION_LOCATION_CALLABLE, EarthProtocol
from upstage_des.units.convert import STANDARD_TIMES, TIME_ALTERNATES, unit_convert

CONTEXT_ERROR_MSG = "Undefined context variable: use EnvironmentContext"


if TYPE_CHECKING:
    from upstage_des.actor import Actor
    from upstage_des.resources.monitoring import MonitoringMixin


class DotDict(dict):
    """A dictionary that supports dot notation as well as dictionary access notation.

    Usage: d = DotDict({'val1':'first'})
    set attributes: d.val2 = 'second' or d['val2'] = 'second'
    get attributes: d.val2 or d['val2'] would both produce 'second'
    """

    def __getattr__(self, key: str) -> Any:
        """Getattr with error for stage.

        Args:
            key (str): The key

        Returns:
            Any: The value
        """
        if key not in self:
            raise AttributeError(f"No key `{key}` found in stage. Use `UP.add_stage_variable`")
        return self.get(key)

    def __setattr__(self, key: str, value: Any) -> None:
        """Set the attribute.

        Typing is upset at a simple pass-through.

        Args:
            key (str): Key
            value (Any): Value
        """
        if key in self:
            raise AttributeError(f"Key {key} is already set.")
        self.__setitem__(key, value)

    def __delattr__(self, key: str) -> None:
        """Delete an attribute.

        Args:
            key (str): Key
        """
        del self[key]


class StageProtocol(Protocol):
    """Protocol for typing the minimum entries in the Stage."""

    @property
    def altitude_units(self) -> str:
        """Units of altitude."""

    @property
    def distance_units(self) -> str:
        """Units of distance."""

    @property
    def stage_model(self) -> EarthProtocol:
        """Model for geodetics."""

    @property
    def intersection_model(self) -> INTERSECTION_LOCATION_CALLABLE:
        """Callable for geodetic intersections."""

    @property
    def time_unit(self) -> str:
        """Time unit, Treated as 'hr' if not set.

        This value modifies ``pretty_now`` from ``UpstageBase``,
        and can be used to modfy ``Wait`` timeouts.
        """

    @property
    def random(self) -> Random:
        """Random number generator."""

    @property
    def daily_time_count(self) -> float | int:
        """The number of time_units in a "day".

        This value only modifies ``pretty_now`` from ``UpstageBase``.

        This is only used if the time_unit is not
        s, min, or hr. In that case, 24 hour days are
        assumed.
        """

    @property
    def debug_log_time(self) -> bool:
        """Whether or not times are logged as a string in the debug logs.

        Can be modified at the individual actor level with debug_log_time.

        Returns:
            bool: If time is logged.
        """

    if TYPE_CHECKING:

        def __getattr__(self, key: str) -> Any: ...

        def __setattr__(self, key: str, value: Any) -> None: ...

        def __delattr__(self, key: str) -> None: ...


class UpstageError(Exception):
    """Raised when an UPSTAGE error happens or expectation is not met."""


class SimulationError(UpstageError):
    """Raised when a simulation error occurs."""

    def __init__(self, message: str, time: float | None = None):
        """Create an informative simulation error.

        Args:
            message (str): Error message
            time (float | None, optional): Time of the error. Defaults to None.
        """
        msg = "Error in the simulation: "
        if msg in message:
            msg = ""
        msg += f" at time {time}: " if time is not None else ""
        self.message = msg + message
        super().__init__(self.message)


class MotionAndDetectionError(SimulationError):
    """A simulation error raised during motion detection."""


class RulesError(UpstageError):
    """Raised by the user when a simulation rule is violated."""


class MockEnvironment:
    """A fake environment that holds the ``now`` property and all-caps attributes."""

    def __init__(self, now: float):
        """Create the mock environment.

        Args:
            now (float): The time the environment is at.
        """
        self.now = now

    @classmethod
    def mock(cls, env: Union[SimpyEnv, "MockEnvironment"]) -> "MockEnvironment":
        """Create a mock environment from another environment.

        Args:
            env (SimpyEnv | MockedEnvironment): The simpy environments

        Returns:
            MockEnvironment: The mocked environment (time only)
        """
        mock_env = cls(now=env.now)
        # copy over any attributes if they are all-caps
        for k, v in env.__dict__.items():
            if k.upper() == k and not k.startswith("_"):
                setattr(mock_env, k, v)
        return mock_env

    @classmethod
    def run(cls, until: float | int) -> None:
        """Method stub for playing nice with rehearsal.

        Args:
            until (float | int): Placeholder
        """
        raise UpstageError("You tried to use `run` on a mock environment")


@dataclass
class SpecialContexts:
    """Accessible lists of typed objects for contexts."""

    actors: list["Actor"] = field(default_factory=list)
    monitored: list["MonitoringMixin"] = field(default_factory=list)


ENV_CONTEXT_VAR: ContextVar[SimpyEnv] = ContextVar("Environment")
SPECIAL_ENTITY_CONTEXT_VAR: ContextVar[SpecialContexts] = ContextVar("SpecialContexts")
ENTITY_CONTEXT_VAR: ContextVar[dict[str, list["NamedUpstageEntity"]]] = ContextVar("Entities")
STAGE_CONTEXT_VAR: ContextVar[DotDict] = ContextVar("Stage")


SKIP_GROUPS: list[str] = ["Actor", "Task", "Location", "CartesianLocation", "GeodeticLocation"]


class UpstageBase:
    """A base mixin class for everyone.

    Provides access to all context variables created by `EnvironmentContext`.

    >>> with EnvironmentContext(initial_time=0.0) as env:
    >>>     data = UpstageBase()
    >>>     assert data.env is env
    """

    def __init__(self) -> None:
        """Simple init to check if environment should be set."""
        try:
            _ = self.env
        except UpstageError:
            warn(f"Environment not created at instantiation of {self}")

    @property
    def env(self) -> SimpyEnv:
        """Return the environment.

        Returns:
            SimpyEnv: SimPy environment.
        """
        try:
            env: SimpyEnv = ENV_CONTEXT_VAR.get()
        except LookupError:
            raise UpstageError("No environment found or set.")
        return env

    @property
    def stage(self) -> StageProtocol:
        """Return the stage context variable.

        Returns:
            StageProtocol: The stage, as defined in context.
        """
        try:
            stage = STAGE_CONTEXT_VAR.get()
        except LookupError:
            raise UpstageError("No stage found or set.")
        return stage

    def get_actors(self) -> list["Actor"]:
        """Return all actors that the director knows.

        Returns:
            list[NamedUpstageEntity]: List of actors in the simulation.
        """
        ans: list[Actor] = []
        try:
            ans = SPECIAL_ENTITY_CONTEXT_VAR.get().actors
        except LookupError:
            raise UpstageError(CONTEXT_ERROR_MSG)
        return ans

    def get_entity_group(self, group_name: str) -> list["NamedUpstageEntity"]:
        """Get a single entity group by name.

        Args:
            group_name (str): The name of the entity group.

        Returns:
            list[NamedUpstageEntity]: List of entities in the group.
        """
        ans: list[NamedUpstageEntity] = []
        try:
            grps: dict[str, list[NamedUpstageEntity]] = ENTITY_CONTEXT_VAR.get()
            ans = grps.get(group_name, [])
        except LookupError:
            raise UpstageError(CONTEXT_ERROR_MSG)
        return ans

    def get_monitored(self) -> list["MonitoringMixin"]:
        """Return entities that inherit from the MonitoringMixin.

        Returns:
            list[MonitoringMixin]: List of entitites that are monitoring.
        """
        ans: list[MonitoringMixin] = []
        try:
            ans = SPECIAL_ENTITY_CONTEXT_VAR.get().monitored
        except LookupError:
            raise UpstageError(CONTEXT_ERROR_MSG)
        return ans

    def get_all_entity_groups(self) -> dict[str, list["NamedUpstageEntity"]]:
        """Get all entity groups.

        Returns:
            dict[str, list[NamedUpstageEntity]]: Entity group names and associated
                entities.
        """
        grps: dict[str, list[NamedUpstageEntity]] = {}
        try:
            grps = ENTITY_CONTEXT_VAR.get()
        except LookupError:
            raise UpstageError(CONTEXT_ERROR_MSG)
        return grps

    @property
    def pretty_now(self) -> str:
        """A well-formatted string of the sim time.

        Tries to account for generic names for time, such as 'ticks'.

        Returns:
            str: The sim time
        """
        now = self.env.now
        time_unit = self.stage.get("time_unit", None)
        # If it's explicitly set to None, still treat it as hours.
        time_unit = "hr" if time_unit is None else time_unit
        standard = TIME_ALTERNATES.get(time_unit.lower(), time_unit)

        ts: str
        if standard in STANDARD_TIMES:
            now_hrs = unit_convert(now, time_unit, "hr")
            day = floor(now_hrs / 24)
            rem_hours = now_hrs - (day * 24)
            hms = strftime("%H:%M:%S", gmtime(rem_hours * 3600))
            ts = f"[Day {day:4.0f} - {hms:s}]"
        else:
            day_unit_count = self.stage.get("daily_time_count", None)
            if day_unit_count is None:
                ts = f"[{now:.3f} {time_unit}]"
            else:
                days = int(floor(now / day_unit_count))
                rem = now - (days * day_unit_count)
                ts = f"[Day {days:4d} - {rem:.3f} {time_unit}]"

        return ts


class NamedUpstageEntity(UpstageBase):
    """A base class for naming entities, and retrieving them.

    This creates a record of every instance of a subclass of this class.

    Example:
        >>> class RocketCar(NamedUpstageEntity, entity_groups=["car", "fast"])
        >>>     ...
        >>> rc = RocketCar()
        >>> assert rc in rc.get_entity_group("car")
    """

    @classmethod
    def __init_subclass__(
        cls,
        entity_groups: Iterable[str] | str | None = None,
        add_to_entity_groups: bool = True,
        skip_classname: bool = False,
    ) -> None:
        if not add_to_entity_groups:
            return

        entity_groups = [] if entity_groups is None else entity_groups

        if isinstance(entity_groups, str):
            entity_groups = [entity_groups]

        entity_groups = list(entity_groups)

        if cls.__name__ not in entity_groups and not skip_classname:
            entity_groups.append(cls.__name__)

        entity_group = list(set(entity_groups))
        old_init = cls.__init__

        @wraps(old_init)
        def the_actual_init(inst: NamedUpstageEntity, *args: Any, **kwargs: Any) -> None:
            inst._add_entity(entity_group)
            old_init(inst, *args, **kwargs)

        setattr(cls, "__init__", the_actual_init)

    def _add_to_group(self, group_name: str) -> None:
        """Add to a single group.

        Args:
            group_name (str): Group name
        """
        try:
            ans = ENTITY_CONTEXT_VAR.get()
            ans.setdefault(group_name, [])
            if self in ans[group_name]:
                raise UpstageError(f"Entity: {self} already recorded in the environment")
            ans[group_name].append(self)
        except LookupError:
            entity_groups = {group_name: [self]}
            ENTITY_CONTEXT_VAR.set(entity_groups)

    def _add_special_group(self) -> None:
        """Add to a special group.

        Sub-classable for type help.

        Make sure that whatever entity group name this goes to is in SKIP_GROUPS.
        """
        ...

    def _add_entity(self, group_names: list[str]) -> None:
        """Add self to an entity group(s).

        Args:
            group_names (list[str]): Group names to add to
        """
        for group_name in group_names:
            if group_name in SKIP_GROUPS:
                continue
            self._add_to_group(group_name)
        self._add_special_group()


class SettableEnv(UpstageBase):
    """A mixin class for allowing the instance's environment to change."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Passthrough for the mixed classes."""
        self._new_env: MockEnvironment | None = None
        super().__init__(*args, **kwargs)

    @property  # type: ignore [override]
    def env(self) -> SimpyEnv | MockEnvironment:
        """Get the relevant environment.

        Returns:
            SimpyEnv | MockEnvironment: Real or mocked environment.
        """
        if self._new_env is not None:
            return self._new_env
        return super().env

    @env.setter
    def env(self, value: MockEnvironment) -> None:
        if isinstance(value, MockEnvironment):
            self._new_env = value
        else:
            # otherwise set new env back to none
            self._new_env = None


class EnvironmentContext:
    """A context manager to create a safe, globally (in context) referenceable environment and data.

    The environment created is of type simpy.Environment

    This also sets context variables for actors, entities, and the stage.

    Usage:
        >>> with EnvironmentContext(initial_time=0.0) as env:
        >>>    env.run(until=3.0)

    This context manager is meant to be paired with inheritors of `UpstageBase`.

    that provides access to the context variables created in this manager.

    >>> class SimData(UpstageBase):
    >>>     ...
    >>>
    >>> with EnvironmentContext(initial_time=0.0) as env:
    >>>     data = SimData()
    >>>     assert data.env is env

    You may also provide a random seed, and a default Random() will be created with
    that seed.

    >>> with EnvironmentContext(random_seed=1234986) as env:
    >>>    UpstageBase().stage.random.uniform(1, 3)
    ...    2.348057489610457

    Or your own RNG:

    >>> rng = Random(1234986)
    >>> with EnvironmentContext(random_gen=rng) as env:
    >>>    UpstageBase().stage.random.uniform(1, 3)
    ...    2.348057489610457
    """

    def __init__(
        self,
        initial_time: float = 0.0,
        random_seed: int | None = None,
        random_gen: Any | None = None,
    ) -> None:
        """Create an environment context.

        random_seed is ignored if random_gen is given. Otherwise random.Random is
        used.

        Args:
            initial_time (float, optional): Time to start the clock at. Defaults to 0.0.
            random_seed (int | None, optional): Seed for RNG. Defaults to None.
            random_gen (Any | None, optional): RNG object. Defaults to None.
        """
        self.env_ctx = ENV_CONTEXT_VAR
        self.special_ctx = SPECIAL_ENTITY_CONTEXT_VAR
        self.entity_ctx = ENTITY_CONTEXT_VAR
        self.stage_ctx = STAGE_CONTEXT_VAR
        self.env_token: Token[SimpyEnv]
        self.special_token: Token[SpecialContexts]
        self.entity_token: Token[dict[str, list[NamedUpstageEntity]]]
        self.stage_token: Token[DotDict]
        self._env: SimpyEnv | None = None
        self._initial_time: float = initial_time
        self._random_seed: int | None = random_seed
        self._random_gen: Any = random_gen

    def __enter__(self) -> SimpyEnv:
        """Create the environment context.

        Returns:
            SimpyEnv: Simpy Environment
        """
        self._env = SimpyEnv(initial_time=self._initial_time)
        self.env_token = self.env_ctx.set(self._env)
        self.special_token = self.special_ctx.set(SpecialContexts())
        self.entity_token = self.entity_ctx.set(defaultdict(list))
        stage = DotDict()
        self.stage_token = self.stage_ctx.set(stage)
        if self._random_gen is None:
            random = Random(self._random_seed)
            stage.random = random
        else:
            stage.random = self._random_gen
        return self._env

    def __exit__(self, *_: Any) -> None:
        """Leave the context."""
        self.env_ctx.reset(self.env_token)
        self.special_ctx.reset(self.special_token)
        self.entity_ctx.reset(self.entity_token)
        self.stage_ctx.reset(self.stage_token)
        self._env = None


def add_stage_variable(varname: str, value: Any) -> None:
    """Add a variable to the stage.

    Will fail if it already exists.

    Args:
        varname (str): Name of the variable
        value (Any): Value to set it as
    """
    try:
        stage = STAGE_CONTEXT_VAR.get()
    except LookupError:
        raise ValueError("Stage should have been set.")
    if varname in stage:
        raise UpstageError(f"Variable '{varname}' already exists in the stage")
    setattr(stage, varname, value)


def get_stage_variable(varname: str) -> Any:
    """Get a variable from the context's stage.

    Args:
        varname (str): Name of the variable

    Returns:
        Any: The variable's value
    """
    try:
        stage = STAGE_CONTEXT_VAR.get()
    except LookupError:
        raise ValueError("Stage should have been set.")
    if varname not in stage:
        raise UpstageError(f"Variable '{varname}' does not exist in the stage")
    return getattr(stage, varname)


def get_stage() -> StageProtocol:
    """Return the entire stage object.

    Returns:
        StageProtocol: The stage
    """
    try:
        stage = STAGE_CONTEXT_VAR.get()
    except LookupError:
        raise ValueError("Stage should have been set.")
    return stage


def create_top_context(
    initial_time: float = 0.0,
    random_seed: int | None = None,
    random_gen: Any | None = None,
) -> EnvironmentContext:
    """Create a stage at this level of context.

    Makes your current level the same as the context manager.

    Returns:
        EnvironmentContext: The context
    """
    ctx = EnvironmentContext(initial_time, random_seed, random_gen)
    ctx.__enter__()
    return ctx


def clear_top_context(ctx: EnvironmentContext) -> None:
    """Clear the context.

    Args:
        ctx (EnvironmentContext): The object made from create_stage()
    """
    ctx.__exit__()



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

import simpy as SIM

import upstage_des.api as UP
from upstage_des.type_help import SIMPY_GEN, TASK_GEN


class CPU(UP.Actor):
    n_procs = UP.State[int](default=0, valid_types=int, recording=True)
    jobs = UP.ResourceState[SIM.Store](default=SIM.Store)

    @staticmethod
    def time_from_data(process_data: dict[str, float]) -> float:
        best_time = process_data["best time"]
        percent_complete = process_data["percent complete"]
        alloc = process_data["allocated"]
        work_left = best_time * (1 - percent_complete)
        time = work_left / alloc
        return time

    @staticmethod
    def left_from_partial(
        process_data: dict[str, float], start_time: float, curr_time: float
    ) -> float:
        alloc = process_data["allocated"]
        work_left = process_data["best time"] * (1 - process_data["percent complete"])
        elapsed = curr_time - start_time
        percent_of_left = 1 - (elapsed / (work_left / alloc))
        percent_left = (1 - process_data["percent complete"]) * percent_of_left
        return percent_left


class CPUProcessStart(UP.DecisionTask):
    def get_name(self) -> str:
        return f"{self._network_name}"

    def make_decision(self, *, actor: CPU) -> None:
        knowledge_name = self.get_name()
        process_data: dict[str, float] = self.get_actor_knowledge(
            actor, knowledge_name, must_exist=True
        )
        # the task takes some amount of time to finish based on its cpu amount
        assert process_data["percent complete"] == 0.0
        # Touch the nucleus variable before it affects this network
        actor.n_procs += 1
        # Now we can add ourselves to the nucleus
        nucleus = actor.get_nucleus()
        nucleus.add_network(knowledge_name, ["n_procs"])


class CPUProcess(UP.Task):
    def get_name(self) -> str:
        return f"{self._network_name}"

    def task(self, *, actor: CPU) -> TASK_GEN:
        knowledge_name = self.get_name()
        process_data: dict[str, float] = self.get_actor_knowledge(
            actor, knowledge_name, must_exist=True
        )
        # We know at this point we're part of the n_procs
        allocate_amount = 1 / (actor.n_procs)
        process_data["allocated"] = allocate_amount
        self.set_actor_knowledge(actor, knowledge_name, process_data, overwrite=True)
        self.set_marker("RUNNING")
        time = actor.time_from_data(process_data)
        print(
            f"{self.env.now:.2f}: Starting: {knowledge_name}\n\tAllocated: {allocate_amount:.2f}"
            f"\n\tTime Left: {time:.2f}"
        )
        yield UP.Wait(time)
        self.clear_actor_knowledge(actor, knowledge_name)
        actor.get_nucleus().remove_network(self.get_name())
        actor.n_procs -= 1
        print(f"{self.env.now:.2f}: Done with: {knowledge_name}")

    def on_interrupt(self, *, actor: CPU, cause: str | UP.NucleusInterrupt) -> UP.InterruptStates:
        if isinstance(cause, UP.NucleusInterrupt):
            assert cause.state_name == "n_procs"

            start_time = self.get_marker_time()
            assert start_time is not None
            knowledge_name = self.get_name()
            process_data: dict[str, float] = self.get_actor_knowledge(
                actor, knowledge_name, must_exist=True
            )
            perc = actor.left_from_partial(process_data, start_time, self.env.now)
            process_data["percent complete"] = perc
            self.set_actor_knowledge(actor, knowledge_name, process_data, overwrite=True)

            return self.INTERRUPT.RESTART
        raise UP.SimulationError("Unexpected interrupt state")


cpu_job_factory = UP.TaskNetworkFactory.from_ordered_terminating(
    name="SingleJob", task_classes=[CPUProcessStart, CPUProcess]
)


class CPUJobFarmer(UP.Task):
    def task(self, *, actor: CPU) -> TASK_GEN:
        job = yield UP.Get(actor.jobs)

        suggest = actor.suggest_network_name(cpu_job_factory)
        new_net = cpu_job_factory.make_network(other_name=suggest)
        actor.add_task_network(new_net)

        proc_know = {"best time": job, "percent complete": 0.0}
        self.set_actor_knowledge(actor, suggest, proc_know)
        actor.start_network_loop(suggest, init_task_name="CPUProcessStart")


cpu_farmer_factory = UP.TaskNetworkFactory.from_single_looping(
    name="Dispatch", task_class=CPUJobFarmer
)


def test_nucleus_sharing() -> None:
    job_time_list = [100.0, 10.0, 15.0, 13.0, 5.0, 25.0]
    job_start_delay = [0.0, 3.0, 5.0, 10.0, 10.0, 3.0]

    with UP.EnvironmentContext() as env:
        cpu = CPU(
            name="Magic Computer",
            n_procs=0,
        )
        _ = UP.TaskNetworkNucleus(actor=cpu)

        net = cpu_farmer_factory.make_network()
        cpu.add_task_network(net)
        cpu.start_network_loop(net.name, "CPUJobFarmer")

        def job_sender() -> SIMPY_GEN:
            for time, delay in zip(job_time_list, job_start_delay):
                yield env.timeout(delay)
                yield cpu.jobs.put(time)

        env.process(job_sender())
        env.run()
        print(env.now)



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

import pytest

from upstage_des.geography import WGS84


def test_distance_and_bearing(atl: tuple[float, float], nas: tuple[float, float]) -> None:
    dist = WGS84.distance(atl, nas, "nmi")
    assert dist == pytest.approx(186.64143600171283)
    bearing = WGS84.bearing(atl, nas)
    assert bearing == pytest.approx(321.4499587022779)
    d2, b2 = WGS84.distance_and_bearing(atl, atl)
    assert pytest.approx(0) == d2


def test_linspace(atl: tuple[float, float], nas: tuple[float, float]) -> None:
    latlons = WGS84.geo_linspace(atl, nas, 10)
    assert len(latlons) == 11
    lats, lons = zip(*latlons)
    assert lats[0] == atl[0]
    assert lats[-1] == pytest.approx(nas[0])
    assert lons[0] == pytest.approx(atl[1])
    assert lons[-1] == pytest.approx(nas[1])


def test_geo_circle(atl: tuple[float, float]) -> None:
    latlons = WGS84.geo_circle(atl, radius=100, num_points=20)
    for lat, lon in latlons:
        d = WGS84.distance(atl, (lat, lon))
        assert d == pytest.approx(100, rel=0.001)


def test_bearing_dist_from_point(atl: tuple[float, float], nas: tuple[float, float]) -> None:
    dist, bearing = WGS84.distance_and_bearing(atl, nas, units="km")
    pt = WGS84.point_from_bearing_dist(atl, bearing, dist, distance_units="km")
    new_dist = WGS84.distance(pt, nas, units="km")
    assert 0 == pytest.approx(new_dist, abs=0.01)


def test_addtional_2() -> None:
    point_1 = (50 + 3 / 60 + 58.76 / 3600, -(5 + 42 / 60 + 53.10 / 3600))
    point_2 = (58 + 38 / 60 + 38.48 / 3600, -(3 + 4 / 60 + 12.34 / 3600))
    dist, bearing = WGS84.distance_and_bearing(point_1, point_2, units="km")
    assert dist == pytest.approx(969.954114)
    actual_bearing = 9 + 8 / 60 + 30.70 / 3600
    assert bearing == pytest.approx(actual_bearing)


def test_addtional_3() -> None:
    point_1 = (-(37 + 57 / 60 + 3.72030 / 3600), 144 + 25 / 60 + 29.52440 / 3600)
    point_2 = (-(37 + 39 / 60 + 10.15610 / 3600), 143 + 55 / 60 + 35.38390 / 3600)
    dist, bearing = WGS84.distance_and_bearing(point_1, point_2, units="km")
    assert dist == pytest.approx(54.972271)
    actual_bearing = 306 + 52 / 60 + 5.37 / 3600
    assert bearing == pytest.approx(actual_bearing)



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

import pytest

from upstage_des.geography import conversions, spherical, wgs84
from upstage_des.geography.conversions import BaseConversions

SC = conversions.SphericalConversions
WSGC = conversions.WGS84Conversions
SC2 = spherical.Spherical
WSGC2 = wgs84.WGS84


@pytest.mark.parametrize("use", [SC, SC2, WSGC, WSGC2])
def test_conversions(use: BaseConversions, random_lla: list[tuple[float, float, float]]) -> None:
    # Do a back and forth test of random Lat Lon Alt
    ecef = use.lla2ecef(random_lla)
    lla_from_ecef = use.ecef2lla(ecef)
    for a, b in zip(lla_from_ecef, random_lla):
        assert pytest.approx(a) == b



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""This file contains a motion manager that does time-stepping."""

from collections.abc import Callable, Generator
from typing import Any, cast

from simpy import Event as SimpyEvent

from upstage_des.actor import Actor
from upstage_des.base import SimulationError, UpstageBase
from upstage_des.motion.motion import LOC_TYPES, SensorType
from upstage_des.states import CartesianLocationChangingState, GeodeticLocationChangingState
from upstage_des.task import process


class SteppedMotionManager(UpstageBase):
    """Tests relative distances of objects with a location property.

    Reports to "sensor" objects when something enters or exits a range.

    Use this manager when the sensing entities are not static. If they are
    static, use `SensorMotionManager`.

    Detectable objects and sensor objects must have an attribute that is a GeodeticLocationState
    OR CartesianLocationState

    Detectable objects, if they aren't Actors, could implement _get_detection_state() -> bool:`
    to allow this class to ignore them sometimes. The default way is to use a `DetectabilityState`
    on the actor.

    Sensor objects MUST implement these two methods:
    1. `entity_entered_range(object)`
    2. `entity_exited_range(object)`

    The first is called when an entity enters the sensor's visiblity.
    The second is called when an entity leaves the visibility or becomes undetectable.

    The sensor object CAN implement a method called `detection_checker`.
    That method takes the location of an object to detect and returns True/False.

    The motion manager will learn about sensor objects with:

    sensor_motion_manager.add_sensor(sensor_object, radius)

    Where radius is a distance in the units defined in upstage.STAGE.

    Simple usage:
    >>> manager = SteppedMotionManager(timestep=0.1)
    >>> UP.STAGE.motion_manager = manager
    >>>  ...
    >>> manager.add_sensor(binoculars, 'vision_radius')
    >>> manager.add_detectable(bird, 'location')

    # TODO: Unify sensor and movable
    # TODO: Having only moving things be detectable/using `_start_mover`
    is easy, but this class lets us do static detection easier, so we may
    have to go about it differently.
    # TODO: Data structures for efficient distances
    """

    def __init__(self, timestep: float, max_empty_events: int = 3, debug: bool = False) -> None:
        """Create the Stepped motion manager.

        Args:
            timestep (float): Timestep to do all pairs distance checks.
            max_empty_events (int, optional): How many timesteps where no events causes a shutdown.
                Defaults to 3.
            debug (bool, optional): Record data or not. Defaults to False.
        """
        super().__init__()
        self._sensors: dict[SensorType, tuple[Callable[[], float], Callable[[], LOC_TYPES]]] = {}
        self._detectables: dict[Actor, Callable[[], LOC_TYPES]] = {}
        self._in_view: set[tuple[SensorType, Actor]] = set()
        self._timestep = timestep
        self._max_empty_events = max_empty_events
        self._debug = debug
        self._debug_log: list[Any] = []
        self._is_running = False

    def _do_log(self, msg: Any) -> None:
        """Write to a log list.

        Args:
            msg (Any): Anything to append.
        """
        if self._debug:
            self._debug_log.append(msg)

    def _update_awareness(self, sensor: SensorType, object: Actor, visible: bool) -> None:
        """Modify sensor/object awareness.

        Args:
            sensor (SensorType): Sensor
            object (Actor): The sensed
            visible (bool): If the sensed is visible.
        """
        if visible:
            if (sensor, object) not in self._in_view:
                self._in_view.add((sensor, object))
                sensor.entity_entered_range(object)
        else:
            if (sensor, object) in self._in_view:
                self._in_view.remove((sensor, object))
                sensor.entity_exited_range(object)

    def _test_detect(self, detectable: Actor) -> bool:
        """Is an actor detectable?

        Args:
            detectable (Actor): The detectable

        Returns:
            bool: If it can be detected
        """
        if not hasattr(detectable, "_get_detection_state"):
            return True
        detect_state = detectable._get_detection_state()
        if detect_state is None:
            return True
        visibility: bool = getattr(detectable, detect_state)
        return visibility

    @staticmethod
    def _detect_dist(loc1: LOC_TYPES, radius: float, loc2: LOC_TYPES, sensor: SensorType) -> bool:
        """Run a detectability check, including sensor custom function.

        Args:
            loc1 (LOC_TYPES): Sensor location
            radius (float): Sensor radius
            loc2 (LOC_TYPES): Target location
            sensor (SensorType): Sensor object

        Returns:
            bool: If it's detectable
        """
        if hasattr(sensor, "detection_checker"):
            visible = sensor.detection_checker(loc2)
        else:
            dist = loc1.straight_line_distance(loc2)
            visible = dist <= radius
        return cast(bool, visible)

    def _run_detectable(
        self,
        sensor_req: list[SensorType] | None = None,
        detectable_req: list[Actor] | None = None,
    ) -> None:
        """All pairs distance checking.

        Args:
            sensor_req (list[SensorType] | None, optional): Sensors. Defaults to None.
            detectable_req (list[Actor] | None, optional): Detectables. Defaults to None.
        """
        sensor_req = list(self._sensors) if sensor_req is None else sensor_req
        sensor_radii = [self._sensors[s][0]() for s in sensor_req]
        sensor_locs = [self._sensors[s][1]() for s in sensor_req]

        detectable_req = list(self._detectables) if detectable_req is None else detectable_req
        detectable_req = [d for d in detectable_req if self._test_detect(d)]
        detect_locs = [self._detectables[d]() for d in detectable_req]

        for sensor, radius, loc in zip(sensor_req, sensor_radii, sensor_locs):
            for detectable, d_loc in zip(detectable_req, detect_locs):
                if detectable is sensor:
                    continue
                visible = self._detect_dist(loc, radius, d_loc, sensor)
                self._do_log((self.env.now, sensor, loc, detectable, d_loc))
                self._update_awareness(sensor, detectable, visible)

    def _only_event_test(self) -> bool:
        """Determine if there are no events in the queue."""
        if len(self.env._queue) == 0:
            return True
        return False

    @process
    def run(self) -> Generator[SimpyEvent, None, None]:
        """Run the main stepped motion loop.

        Yields:
            Generator[SimpyEvent, None, None]: _description_
        """
        if self._is_running:
            # If run() is called later than a task is queued,
            # then this may be true already.
            return
        self._is_running = True
        n_empty = 0
        while True:
            timeout = self.env.timeout(self._timestep)
            yield timeout
            self._run_detectable()
            # prevents infinite simulations
            if self._only_event_test():
                n_empty += 1
                if n_empty >= self._max_empty_events:
                    return
            else:
                n_empty = 0

    @process
    def run_particular(self, rate: float, detectable: Actor) -> Generator[SimpyEvent, None, None]:
        """Run detections against a single target at a faster rate.

        Args:
            rate (float): Time rate to do detection checks.
            detectable (Actor): The actor to be detected by the known sensors.
        """
        while True:
            yield self.env.timeout(rate)
            self._run_detectable(sensor_req=None, detectable_req=[detectable])

    def add_sensor(
        self,
        sensor: SensorType,
        radius_attr_name: str = "radius",
        location_attr_name: str = "location",
    ) -> None:
        """Add a sensor the motion manager.

        Args:
            sensor (SensorType): The sensing object
            radius_attr_name (str): Radius attribute name. Defaults to "radius".
            location_attr_name (str): Location attribute name. Defaults to "location".
        """
        # test the sensor for earlier errors about improperly-defined methods
        required_methods = ["entity_entered_range", "entity_exited_range"]
        required_attrs = [radius_attr_name, location_attr_name]
        for req in required_methods:
            if not hasattr(sensor, req):
                raise NotImplementedError(f"Sensor {sensor} does not have '{req}' method!")
        for attr in required_attrs:
            if not hasattr(sensor, attr):
                raise SimulationError(f"Sensor {sensor} doesn't have attribute {attr}")

        def get_radius() -> float:
            return cast(float, getattr(sensor, radius_attr_name))

        def get_location() -> LOC_TYPES:
            return cast(LOC_TYPES, getattr(sensor, location_attr_name))

        self._sensors[sensor] = (get_radius, get_location)

    def add_detectable(
        self,
        detectable: Actor,
        location_attr_name: str = "location",
        new_rate: float | None = None,
    ) -> None:
        """Add an object that is detectable to the manager.

        The object must have an attribute that performs distance calculations.
        See the class docstring for more.

        Args:
            detectable (Actor): An object that has a location attribute
            location_attr_name (str): Name of the location attribute. Defaults to "location".
            new_rate (float | None): Optional new rate for a detectable
            (if it needs faster, most likely)
        """
        if not hasattr(detectable, location_attr_name):
            raise SimulationError(
                f"Detectable {detectable} doesn't have attribute {location_attr_name}"
            )
        try:
            self._test_detect(detectable)
        except Exception:
            raise SimulationError(f"Detectable {detectable} needs a detectable state.")

        def get_location() -> LOC_TYPES:
            return cast(LOC_TYPES, getattr(detectable, location_attr_name))

        self._detectables[detectable] = get_location

    def _mover_not_detectable(self, detectable: Actor) -> None:
        """Called via DetectabilityState state when an object becomes undetectable.

        Could be called for any reason; use this feature to alert sensors that
        an object should no longer be considered by that sensor.
        """
        # TODO: key on detectable may make it faster
        to_rem = set()
        for sensor, detect in self._in_view:
            if detect is detectable:
                sensor.entity_exited_range(detectable)
                to_rem.add((sensor, detect))
        self._in_view -= to_rem

    def _mover_became_detectable(self, detectable: Actor) -> None:
        """Called via DetectabilityState state when an object becomes detectable.

        Could be called for any reason; use this feature to alert sensors that
        an object should be considered by that sensor.
        """
        self._run_detectable(detectable_req=[detectable])

    def _start_mover(self, mover: Actor, speed: float, waypoints: list[LOC_TYPES]) -> None:
        # we don't need this method, except to hook into motion states
        if not self._is_running:
            self.run()
        if mover in self._detectables:
            return
        state_name_1 = mover._get_matching_state(GeodeticLocationChangingState)
        state_name_2 = mover._get_matching_state(CartesianLocationChangingState)

        use_state = state_name_1 or state_name_2
        if use_state is None:
            raise SimulationError(f"Mover {mover} doesn't have a Location state")

        self.add_detectable(mover, use_state)

    def _stop_mover(self, mover: Actor) -> None:
        # we don't need this method, except to hook into motion states
        # It may be useful for ending detections, but the user should
        # handle that themselves
        if mover in self._detectables:
            del self._detectables[mover]



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""This file contains the fundamental Actor class for UPSTAGE."""

from collections import defaultdict
from collections.abc import Callable, Iterable
from copy import copy, deepcopy
from dataclasses import dataclass
from inspect import Parameter, signature
from typing import TYPE_CHECKING, Any, Self

from simpy import Process

from upstage_des.events import Event

from .base import (
    SPECIAL_ENTITY_CONTEXT_VAR,
    MockEnvironment,
    NamedUpstageEntity,
    SettableEnv,
    SimulationError,
    UpstageError,
)
from .data_types import CartesianLocation, GeodeticLocation
from .states import (
    ActiveState,
    CartesianLocationChangingState,
    DetectabilityState,
    GeodeticLocationChangingState,
    ResourceState,
    State,
)
from .task import Task
from .task_network import TaskNetwork, TaskNetworkFactory
from .utils import get_caller_info, get_caller_object

__all__ = ("Actor",)

if TYPE_CHECKING:
    from .nucleus import TaskNetworkNucleus

LOC_STATE = GeodeticLocationChangingState | CartesianLocationChangingState
LOCATIONS = GeodeticLocation | CartesianLocation
LOC_LIST = list[GeodeticLocation] | list[CartesianLocation]


@dataclass
class TaskData:
    name: str
    process: Process


class Actor(SettableEnv, NamedUpstageEntity):
    """Actors perform tasks and are composed of states.

    You can subclass, but do not overwrite __init_subclass__.

    Always super().__init__().

    Parameters
    ----------
    name : str
        The name of the actor.  This is a required attribute.
    debug_log : bool, optional
        Run the debug logger which captures runtime information about the
        actor.
    **states
        Keyword arguments to set the values of the actor's states.

    Raises:
    ------
    ValueError
        If the keys of the states passed as keyword arguments do not match the
        names of the actor's states.

    """

    def __init_states(self, **states: Any) -> None:
        seen = set()
        for state, value in states.items():
            if state in self._state_defs:
                seen.add(state)
                setattr(self, state, value)
            else:
                raise UpstageError(f"Input to {self} was not expected: {state}={value}")
        exist = set(self._state_defs.keys())
        unseen = exist - seen
        for state_name in unseen:
            _state = self._state_defs[state_name]
            if _state.has_default():
                seen.add(state_name)
                _state._set_default(self)
        if len(seen) != len(exist):
            raise UpstageError(
                f"Missing values for states! These states need values: "
                f"{exist - seen} to be specified for '{self.name}'."
            )
        if "log" in seen:
            raise UpstageError("Do not name a state `log`")

    def __init__(
        self,
        *,
        name: str,
        debug_log: bool = True,
        debug_log_time: bool | None = None,
        **states: Any,
    ) -> None:
        """Create an Actor.

        Args:
            name (str): The name of the actor.
            debug_log (bool, optional): Whether to write to debug log. Defaults to True.
            debug_log_time (bool, optional): If time is logged in debug messages.
                Defaults to None (uses Stage value), otherwise local value is used.
            states (Any): Values for each state as kwargs.
        """
        self.name = name
        super().__init__()

        self._active_states: dict[str, dict[str, Any]] = {}
        self._num_clones: int = 0
        self._state_defs: dict[str, State] = getattr(self.__class__, "_state_defs", {})

        self._mimic_states: dict[str, tuple[Actor, str]] = {}  # has to be before other calls
        self._mimic_states_by_task: dict[Task, set[str]] = defaultdict(set)

        self._states_by_task: dict[Task, set[str]] = defaultdict(set)
        self._tasks_by_state: dict[str, set[Task]] = defaultdict(set)

        self._task_networks: dict[str, TaskNetwork] = {}
        self._task_queue: dict[str, list[str]] = {}

        self._knowledge: dict[str, Any] = {}
        self._is_rehearsing: bool = False

        self._debug_logging: bool = debug_log
        self._debug_log_time = debug_log_time
        self._debug_log: list[tuple[float | int, str]] = []

        self._state_histories: dict[str, list[tuple[float, Any]]] = {}

        # Task Network Nucleus hook-ins
        self._state_listener: TaskNetworkNucleus | None = None

        self.__init_states(**states)

    def __init_subclass__(
        cls,
        *args: Any,
        entity_groups: Iterable[str] | str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init_subclass__(entity_groups=entity_groups)
        # get the states
        states = {}
        all_states = {}
        # This ensures that newer classes overwrite older states
        for base_class in cls.mro()[::-1]:
            for state_name, state in base_class.__dict__.items():
                if isinstance(state, State):
                    if base_class == cls:
                        states[state_name] = state
                        state.name = state_name
                    all_states[state_name] = state
        cls._state_defs = all_states

        nxt = cls.mro()[1]
        if nxt is object:
            raise UpstageError(f"Actor has bad inheritance, MRO: {cls.mro()}")

        sig = signature(cls.__init__)
        params = list(sig.parameters.values())
        # Find the "states=" parameter of the signature and remove it.
        state_parameter = [x for x in params if x.name == "states"]
        if state_parameter:
            params.remove(state_parameter[0])
        for state in states:
            params.insert(-1, Parameter(state, Parameter.KEYWORD_ONLY))
        try:
            setattr(cls.__init__, "__signature__", sig.replace(parameters=params))
        except ValueError as e:
            e.add_note("Failure likely due to repeated state name in inherited actor")
            raise e

    def _add_special_group(self) -> None:
        """Add self to the actor context list.

        Called by the NamedUpstageEntity on group inits.
        """
        ans = SPECIAL_ENTITY_CONTEXT_VAR.get().actors
        if self in ans:
            return
        ans.append(self)

    def _lock_state(self, *, state: str, task: Task) -> None:
        """Lock one of the actor's states by a given task.

        Args:
            state (str): The name of the state to lock
            task (Task): The task that is locking the state
        """
        the_state = self._state_defs[state]
        if not the_state.IGNORE_LOCK:
            # single-task only, so no task should
            # be associated with this state
            if self._tasks_by_state[state]:
                raise SimulationError(
                    f"State '{state}' cannot be used by '{task}' because it is "
                    f"locked by {self._tasks_by_state[state]}"
                )
        else:
            # We can have multiple locks, but make sure we are repeating a lock
            if task in self._tasks_by_state[state]:
                raise SimulationError(
                    f"State '{state}' already locked by '{task}'. "
                    "Did you forget to unlock/deactivate it?"
                )
        self._states_by_task[task].add(state)
        self._tasks_by_state[state].add(task)

    def _set_active_state_data(
        self,
        state_name: str,
        started_at: float | None = None,
        **data: Any,
    ) -> None:
        """Set the data for an active state.

        Args:
            state_name (str): Name of the state
            started_at (Optional[float], optional): Time the data is set at. Defaults to None.
            **data (Any): key:values as kwargs for the state data.
        """
        # Rule: underscored active data will get remembered
        started_at = self.env.now if started_at is None else started_at
        old_data = self._active_states.get(state_name, {})
        new_data = {"started_at": started_at, **data}
        keep_old = {k: v for k, v in old_data.items() if k not in new_data and "_" == k[0]}
        new_data.update(keep_old)
        self._active_states[state_name] = new_data

    def activate_state(
        self,
        *,
        state: str,
        task: Task,
        **kwargs: Any,
    ) -> None:
        """Set a state as active.

        Note:
            This method is used by the tasks for activating states they use/modify.

        TODO: on init, create `activate_<statename>` methods that type-hint the inputs

        Args:
            state (str): The name of the state to set as active.
            task (Task): The task that is activating the state.
            **kwargs (Any): key:values as kwargs for the state activation.
        """
        if state not in self._state_defs:
            raise SimulationError(f"No state named '{state}' to activate")
        self._lock_state(state=state, task=task)
        self._set_active_state_data(state_name=state, started_at=self.env.now, task=task, **kwargs)
        # any initialization in the state needs to be called via attribute access
        getattr(self, state)
        # The activation handles getattr
        _state = self._state_defs[state]
        assert isinstance(_state, ActiveState)
        _state.activate(self, task=task)

    def activate_linear_state(self, *, state: str, rate: float, task: Task) -> None:
        """Shortcut for activating a LinearChangingState.

        Args:
            state (str): The name of the LinearChangingState to set as active.
            rate (float): The rate of the change
            task (Task): The task that is activating
        """
        self.activate_state(state=state, task=task, rate=rate)

    def activate_location_state(
        self, *, state: str, speed: float, waypoints: LOC_LIST, task: Task
    ) -> None:
        """Shortcut for activating a (Cartesian|Geodetic)LocationChangingState.

        Args:
            state (str): State name
            speed (float): The speed to move at
            waypoints (LOC_LIST): Waypoints to move over
            task (Task): The task that the state is activated during.
        """
        self.activate_state(
            state=state,
            speed=speed,
            waypoints=waypoints,
            task=task,
        )

    def _unlock_state(self, *, state: str, task: Task) -> None:
        """Release a task's lock of a state.

        Args:
            state (str): The name of the state to lock
            task (Task): The task that is locking the state
        """
        the_state = self._state_defs[state]
        if not the_state.IGNORE_LOCK:
            # single-task only, so only one task should
            # be associated with this state
            if task not in self._tasks_by_state[state]:
                raise SimulationError(
                    f"State `{state}` isn't locked by '{task}', but it's trying to be unlocked."
                )
            self._states_by_task[task].remove(state)
            self._tasks_by_state[state].remove(task)
        elif task in self._tasks_by_state[state]:
            self._states_by_task[task].remove(state)
            self._tasks_by_state[state].remove(task)
        else:
            raise UpstageError(f"State '{state}' was not activated by '{task}', cannot deactivate")

    def deactivate_states(self, *, states: str | Iterable[str], task: Task) -> None:
        """Set a list of active states to not active.

        Args:
            states (str | Iterable[str]): The names of the states to deactivate.
            task (Task): The task that is deactivating the states.
        """
        if isinstance(states, str):
            states = [states]

        for state in states:
            self.deactivate_state(state=state, task=task)

    def deactivate_state(self, *, state: str, task: Task) -> None:
        """Deactivate a specific state.

        Args:
            state (str): The name of the state to deactivate.
            task (Task): The task that is deactivating the state.
        """
        self._unlock_state(state=state, task=task)

        # the deactivated state may need to be updated
        getattr(self, state)
        # and then deactivate it, only if it was unlocked
        the_state = self._state_defs[state]
        if not isinstance(the_state, ActiveState):
            raise UpstageError(f"Stage {state} is not an active type state.")
        ignore = the_state.deactivate(self, task=task)
        if state in self._active_states and not ignore:
            del self._active_states[state]

    def deactivate_all_states(self, *, task: Task) -> None:
        """Deactivate all states in the actor for a given task.

        Args:
            task (Task): The task that is deactivating the states.
        """
        state_names = list(self._states_by_task[task])
        self.deactivate_states(states=state_names, task=task)

    def get_active_state_data(
        self, state_name: str, without_update: bool = False
    ) -> dict[str, Any]:
        """Get the data for a specific state.

        Args:
            state_name (str): The name of the state for which to retrieve the data.
            without_update (bool): Whether or not to update the state to the current
                sim time. Defaults to True

        Returns:
            dict[str, Any]: The state data.
        """
        if not without_update:
            getattr(self, state_name)
        ans: dict[str, Any] = self._active_states.get(state_name, {})
        return ans

    def _mimic_state_name(self, self_state: str) -> str:
        """Create a mimic state name.

        Args:
            self_state (str): The name of the state

        Returns:
            str: Mimic-safe name
        """
        return f"{id(self)}-{self_state}"

    def activate_mimic_state(
        self,
        *,
        self_state: str,
        mimic_state: str,
        mimic_actor: "Actor",
        task: Task,
    ) -> None:
        """Activate a state to mimic a state on another actor.

        Args:
            self_state (str): State name to be the mimic
            mimic_state (str): State on the other actor to be mimiced
            mimic_actor (Actor): The other actor.
            task (Task): The task during which the state is mimiced.
        """
        caller = get_caller_object()
        if isinstance(caller, Task) and caller._rehearsing:
            raise UpstageError("Mimic state activated on rehearsal. This is unsupported/unstable")
        if self_state in self._mimic_states:
            raise UpstageError(f"{self_state} already mimicked")
        self._mimic_states[self_state] = (mimic_actor, mimic_state)
        self._mimic_states_by_task[task].add(self_state)

        state = self._state_defs[self_state]
        self_state_name = self._mimic_state_name(self_state)
        if state.is_recording:

            def recorder(instance: Actor, value: Any) -> None:
                if instance is mimic_actor:
                    state._do_record(self, value)

            mimic_actor._add_callback_to_state(self_state_name, recorder, mimic_state)

    def deactivate_mimic_state(self, *, self_state: str, task: Task) -> None:
        """Deactivate a mimicking state.

        Args:
            self_state (str): State name
            task (Task): Task it's running in.
        """
        getattr(self, self_state)
        mimic_actor, mimic_state = self._mimic_states[self_state]
        state = self._state_defs[self_state]
        self_state_name = self._mimic_state_name(self_state)
        if state.is_recording:
            mimic_actor._remove_callback_from_state(self_state_name, mimic_state)
        del self._mimic_states[self_state]
        self._mimic_states_by_task[task].remove(self_state)

    def deactivate_all_mimic_states(self, *, task: Task) -> None:
        """Deactivate all mimicking states in the actor for a given task.

        Args:
            task (Task): The task where states are mimicking others.
        """
        for state in list(self._mimic_states):
            self.deactivate_mimic_state(self_state=state, task=task)

    def _add_callback_to_state(
        self,
        source: Any,
        callback: Callable[["Actor", Any], Any],
        state_name: str,
    ) -> None:
        """Add a callback to a state for recording.

        Args:
            source (Any): The source for keying the callback (unused, but for the key)
            callback (Callable[[Actor, Any], Any]): Takes the actor and state value
            state_name (str): _description_
        """
        state: State = self._state_defs[state_name]
        state._add_callback(source, callback)

    def _remove_callback_from_state(
        self,
        source: Any,
        state_name: str,
    ) -> None:
        """Remove a state callback based on the source key.

        Args:
            source (Any): Callback key
            state_name (str): Name of the state with the callback.
        """
        state = self._state_defs[state_name]
        state._remove_callback(source)

    def get_knowledge(self, name: str, must_exist: bool = False) -> Any:
        """Get a knowledge value from the actor.

        Args:
            name (str):  The name of the knowledge
            must_exist (bool): Raise an error if the knowledge isn't present. Defaults to false.

        Returns:
            Any: The knowledge value. None if the name doesn't exist.
        """
        if must_exist and name not in self._knowledge:
            raise SimulationError(f"Knowledge '{name}' does not exist in {self}")
        return self._knowledge.get(name, None)

    def get_and_clear_knowledge(self, name: str, caller: str | None = None) -> Any:
        """Get knowledge and clear it.

        Clearing knowledge implies it must exist in the direct methods, so the
        same assumption holds here.

        Args:
            name (str): Knowledge name.
            caller (str): The name of the calling process for logging. Defaults to None.

        Returns:
            Any: The knowledge value.
        """
        know = self.get_knowledge(name, must_exist=True)
        self.clear_knowledge(name, caller)
        return know

    def _log_caller(
        self,
        method_name: str = "",
        caller_level: int = 1,
        caller_name: str | None = None,
    ) -> None:
        """Log information about who is calling this method.

        If no caller_name is given, it is searched for in the stack.

        Args:
            method_name (str, optional): Method name for logging. Defaults to "".
            caller_level (int, optional): Level to look up for the caller. Defaults to 1.
            caller_name (Optional[str], optional): Name of the caller. Defaults to None.
        """
        if caller_name is None:
            info = get_caller_info(caller_level=caller_level + 1)
        else:
            info = caller_name
        self.log(f"method '{method_name}' called by '{info}'")

    def set_knowledge(
        self,
        name: str,
        value: Any,
        overwrite: bool = False,
        caller: str | None = None,
    ) -> None:
        """Set a knowledge value.

        Raises an error if the knowledge exists and overwrite is False.

        Args:
            name (str): The name of the knowledge item.
            value (Any): The value to store for the knowledge.
            overwrite (bool, Optional): Allow the knowledge to be changed if it exits.
                Defaults to False.
            caller (str, Optional): The name of the object that called the method.
        """
        self._log_caller(f"set_knowledge '{name}={value}'", caller_name=caller)
        if name in self._knowledge and not overwrite:
            raise SimulationError(
                f"Actor {self} overwriting existing knowledge {name} "
                f"without override permission. \n"
                f"Current: {self._knowledge[name]}, New: {value}"
            )
        else:
            self._knowledge[name] = value

    def clear_knowledge(self, name: str, caller: str | None = None) -> None:
        """Clear a knowledge value.

        Raises an error if the knowledge does not exist.

        Args:
            name (str): The name of the knowledge item to clear.
            caller (str):  The name of the Task that called the method.
                Used for debug logging purposes.

        """
        self._log_caller(f"clear_knowledge '{name}'", caller_name=caller)
        if name not in self._knowledge:
            raise SimulationError(f"Actor {self} does not have knowledge: {name}")
        else:
            del self._knowledge[name]

    def set_bulk_knowledge(
        self, know: dict[str, Any], overwrite: bool = False, caller: str | None = None
    ) -> None:
        """Set multiple knowledge entries at once.

        Args:
            know (dict[str, Any]): Dictionary of key:value pairs of knowledge.
            overwrite (bool, optional): If overwrite is allowed. Defaults to False.
            caller (str | None, optional): The name of the Task that called the method.
                Defaults to None.
        """
        for k, v in know.items():
            self.set_knowledge(k, v, overwrite, caller)

    def clear_bulk_knowledge(self, names: Iterable[str], caller: str | None = None) -> None:
        """Clear a list of knowledge entries.

        Args:
            names (Iterable[str]): Knowledge names.
            caller (str | None, optional): The name of the Task that called the method.
                Defaults to None.
        """
        for name in names:
            self.clear_knowledge(name, caller)

    def get_bulk_knowledge(self, names: Iterable[str], must_exist: bool = False) -> dict[str, Any]:
        """Get multiple knowledge items.

        Args:
            names (Iterable[str]): Names of the knowledge
            must_exist (bool, optional): If all entires must exist. Defaults to False.

        Returns:
            dict[str, Any]: The knowledge values. None if not present.
        """
        return {name: self.get_knowledge(name, must_exist) for name in names}

    def get_and_clear_bulk_knowledge(
        self, names: Iterable[str], caller: str | None = None
    ) -> dict[str, Any]:
        """Get and clear multiple knowledge entries.

        Args:
            names (Iterable[str]): The knowledge to retrieve and delete.
            caller (str | None, optional): The name of the caller. Defaults to None.

        Returns:
            dict[str, Any]: The retrieved knowledge.
        """
        return {name: self.get_and_clear_knowledge(name, caller) for name in names}

    def add_task_network(self, network: TaskNetwork) -> None:
        """Add a task network to the actor.

        Args:
            network (TaskNetwork): The task network to add to the actor.
        """
        network_name = network.name
        if network_name in self._task_networks:
            raise SimulationError(f"Task network{network_name} already in {self}")
        self._task_networks[network_name] = network
        self._task_queue[network_name] = []

    def clear_task_queue(self, network_name: str) -> None:
        """Empty the actor's task queue.

        This will cause the task network to be used for task flow.

        Args:
            network_name (str): The name of the network to clear the task queue.
        """
        self._log_caller("clear_task_queue")
        self._task_queue[network_name] = []

    def set_task_queue(self, network_name: str, task_list: list[str]) -> None:
        """Initialize an actor's empty task queue.

        Args:
            network_name (str): Task Network name
            task_list (list[str]): List of task names to queue.

        Raises:
            SimulationError: _description_
        """
        self._log_caller("set_task_queue")
        if self._task_queue[network_name]:
            raise SimulationError(f"Task queue on {self.name} is already set. Use append or clear.")
        self._task_queue[network_name] = list(task_list)

    def get_task_queue(self, network_name: str) -> list[str]:
        """Get the actor's task queue on a single network.

        Args:
            network_name (str): The network name

        Returns:
            list[str]: List of task names in the queue
        """
        return self._task_queue[network_name]

    def get_all_task_queues(self) -> dict[str, list[str]]:
        """Get the task queues for all running networks.

        Returns:
            dict[str, list[str]]: Task names, keyed on task network name.
        """
        queues: dict[str, list[str]] = {}
        for name in self._task_networks.keys():
            queues[name] = self.get_task_queue(name)
        return queues

    def get_next_task(self, network_name: str) -> None | str:
        """Return the next task the actor has been told if there is one.

        This does not clear the task, it's information only.

        Args:
            network_name (str): The name of the network

        Returns:
            None | str: The name of the next task, None if no next task.
        """
        queue = self._task_queue[network_name]
        queue_length = len(queue)
        return None if queue_length == 0 else queue[0]

    def _clear_task(self, network_name: str) -> None:
        """Clear a task from the queue.

        Useful for rehearsal.
        """
        self._task_queue[network_name].pop(0)

    def _begin_next_task(self, network_name: str, task_name: str) -> None:
        """Clear the first task in the task queue.

        The task name is required to check that the next task follows the actor's plan.

        Args:
            network_name (str): The task network name
            task_name (str): The name of the task to start
        """
        queue = self._task_queue.get(network_name)
        if queue and queue[0] != task_name:
            raise SimulationError(
                f"Actor {self.name} commanded to perform '{task_name}' but '{queue[0]}' is expected"
            )
        elif not queue:
            self.set_task_queue(network_name, [task_name])
        self.log(f"begin_next_task: Starting {task_name} task")
        self._task_queue[network_name].pop(0)

    def start_network_loop(
        self,
        network_name: str,
        init_task_name: str | None = None,
    ) -> None:
        """Start a task network looping/running on an actor.

        If no task name is given, it will default to following the queue.

        Args:
            network_name (str): Network name.
            init_task_name (str, optional): Task to start with. Defaults to None.
        """
        network = self._task_networks[network_name]
        network.loop(actor=self, init_task_name=init_task_name)

    def get_running_task(self, network_name: str) -> TaskData | None:
        """Return name and process reference of a task on this Actor's task network.

        Useful for finding a process to call interrupt() on.

        Args:
            network_name (str): Network name.

        Returns:
            TaskData: Dataclass of name and process for the current task.
                {"name": Name, "process": the Process simpy is holding.}
        """
        if network_name not in self._task_networks:
            raise SimulationError(f"{self} does not have a task networked named {network_name}")
        net = self._task_networks[network_name]
        if net._current_task_proc is not None:
            assert net._current_task_name is not None
            assert net._current_task_proc is not None
            task_data = TaskData(name=net._current_task_name, process=net._current_task_proc)
            return task_data
        return None

    def get_running_tasks(self) -> dict[str, TaskData]:
        """Get all running task data.

        Returns:
            dict[str, dict[str, TaskData]]: Dictionary of all running tasks.
                Keyed on network name, then {"name": Name, "process": ...}
        """
        tasks: dict[str, TaskData] = {}
        for name, net in self._task_networks.items():
            if net._current_task_proc is not None:
                assert net._current_task_name is not None
                tasks[name] = TaskData(name=net._current_task_name, process=net._current_task_proc)
        return tasks

    def interrupt_network(self, network_name: str, **interrupt_kwargs: Any) -> None:
        """Interrupt a running task network.

        Args:
            network_name (str): The name of the network.
            interrupt_kwargs (Any): kwargs to pass to the interrupt.
        """
        data = self.get_running_task(network_name)
        if data is None:
            raise UpstageError(f"No processes named {network_name} is running.")
        data.process.interrupt(**interrupt_kwargs)

    def has_task_network(self, network_id: Any) -> bool:
        """Test if a network id exists.

        Args:
            network_id (Any): Typically a string for the network name.

        Returns:
            bool: If the task network is on this actor.
        """
        return network_id in self._task_networks

    def suggest_network_name(self, factory: TaskNetworkFactory) -> str:
        """Deconflict names of task networks by suggesting a new name.

        Used for creating multiple parallel task networks.

        Args:
            factory (TaskNetworkFactory): The factory from which you will create the network.

        Returns:
            str: The network name to use
        """
        new_name = factory.name
        if new_name not in self._task_networks:
            return new_name
        i = 0
        while new_name in self._task_networks:
            i += 1
            new_name = f"{factory.name}_{i}"
        return new_name

    def delete_task_network(self, network_id: Any) -> None:
        """Deletes a task network reference.

        Be careful, the network may still be running!

        Do any interruptions on your own.

        Args:
            network_id (Any): Typically a string for the network name.
        """
        if not self.has_task_network(network_id):
            raise SimulationError(f"No networked with id: {network_id} to delete")
        del self._task_networks[network_id]

    def rehearse_network(
        self,
        network_name: str,
        task_name_list: list[str],
        knowledge: dict[str, Any] | None = None,
        end_task: str | None = None,
    ) -> Self:
        """Rehearse a network on this actor.

        Supply the network name, the tasks to rehearse from this state, and
        any knowledge to apply to the cloned actor.

        Args:
            network_name (str): Network name
            task_name_list (list[str]): Tasks to rehearse on the network.
            knowledge (dict[str, Any], optional): knowledge to give to the cloned
                actor. Defaults to None.
            end_task (str, optional): A task to end on once reached.

        Returns:
            Actor: The cloned actor after rehearsing the network.
        """
        knowledge = {} if knowledge is None else knowledge
        net = self._task_networks[network_name]
        understudy = net.rehearse_network(
            actor=self,
            task_name_list=task_name_list,
            knowledge=knowledge,
            end_task=end_task,
        )
        return understudy

    def clone(
        self,
        new_env: MockEnvironment | None = None,
        knowledge: dict[str, Any] | None = None,
        **new_states: Any,
    ) -> Self:
        """Clones an actor and assigns it a new environment.

        Note:
            This function is useful when testing if an actor can accomplish a
            task.

            In general, cloned actor are referred to as ``understudy``
            to keep with the theater analogy.

            The clones' names are appended with the label ``'[CLONE #]'`` where
            ``'#'`` indicates the number of clones of the actor.

        Args:
            new_env (Optional[MockEnvironment], optional): Environment for cloning.
                Defaults to None.
            knowledge (Optional[dict[str, Any]], optional): Knowledge for the clone.
                Defaults to None.
            new_states (Any): New states to add to the actor when cloning.

        Returns:
            Actor: The cloned actor of the same type
        """
        knowledge = {} if knowledge is None else knowledge
        new_env = MockEnvironment.mock(self.env) if new_env is None else new_env

        states: dict[str, Any] = {}
        for state in self.states:
            state_obj = self._state_defs[state]
            if isinstance(state_obj, ResourceState):
                states[state] = state_obj._make_clone(self, getattr(self, state))
            else:
                states[state] = copy(getattr(self, state))
        states.update(new_states)

        self._num_clones += 1

        clone = self.__class__(
            name=self.name + f" [CLONE {self._num_clones}]",
            debug_log=self._debug_logging,
            debug_log_time=self._debug_log_time,
            **states,
        )
        clone.env = new_env

        ignored_attributes = list(states.keys()) + ["env", "stage"]

        for attribute_name, attribute in self.__class__.__dict__.items():
            if not any(
                (
                    attribute_name in ignored_attributes,
                    attribute_name.startswith("_"),
                    callable(attribute),
                )
            ):
                setattr(clone, attribute_name, attribute)

        # update the state histories
        for state_name in self._state_defs:
            if state_name in self._state_histories:
                clone._state_histories[state_name] = deepcopy(self._state_histories[state_name])

        clone._knowledge = {}
        for name, data in self._knowledge.items():
            clone._knowledge[name] = copy(data)

        for name, data in knowledge.items():
            clone._knowledge[name] = copy(data)

        clone._task_queue = copy(self._task_queue)
        clone._task_networks = copy(self._task_networks)

        if clone._debug_logging:
            clone._debug_log = list(self._debug_log)

        clone._is_rehearsing = True
        return clone

    def log(self, msg: str | None = None) -> list[tuple[float | int, str]] | None:
        """Add to the log or return it.

        Only adds to log if debug_logging is True.

        Args:
            msg (str, Optional): The message to log.

        Returns:
            list[str] | None: The log if no message is given. None otherwise.
        """
        if msg is None:
            return self._debug_log
        elif self._debug_logging:
            dlt = self._debug_log_time
            do_time = dlt if dlt is not None else self.stage.get("debug_log_time", True)
            if do_time:
                ts = self.pretty_now
                msg = f"{ts} {msg}"
            self._debug_log += [(self.env.now, msg)]
        return None

    def get_log(self) -> list[tuple[float | int, str]]:
        """Get the debug log.

        Returns:
            list[str]: List of log messages.
        """
        return self._debug_log

    @property
    def states(self) -> tuple[str, ...]:
        """Get the names of the actor's states.

        Returns:
            tuple[str]: State names
        """
        return tuple(self._state_defs.keys())

    @property
    def state_values(self) -> dict[str, Any]:
        """Get the state names and values.

        Returns:
            dict[str, Any]: State name:value pairs.
        """
        return {k: getattr(self, k) for k in self.states}

    def _get_detection_state(self) -> None | str:
        """Find the name of a state is of type DetectabilityState.

        Returns:
            None | str: The name of the state (None if none found).
        """
        detection = [k for k, v in self._state_defs.items() if isinstance(v, DetectabilityState)]
        if len(detection) > 1:
            raise NotImplementedError("Only 1 state of type DetectabilityState allowed for now")
        return None if not detection else detection[0]

    def _match_attr(self, name: str) -> str | None:
        """Test if self has a matching attribute name.

        Args:
            name (str): The attribute to find

        Returns:
            str | None: The name if it has it, None otherwise.
        """
        if not hasattr(self, name):
            return None
        return name

    def _get_matching_state(
        self,
        state_class: type[State],
        attr_matches: dict[str, Any] | None = None,
    ) -> str | None:
        """Find a state that matches the class and optional attributes and return its name.

        For multiple states with the same class, this returns the first available.

        Args:
            state_class (State): The class of state to search for
            attr_matches (Optional[dict[str, Any]], optional): Attributes and values
                to match. Defaults to None.

        Returns:
            str | None: The name of the state (for getattr)
        """

        def match_tester(nm: str, val: Any, state: State) -> bool:
            if hasattr(state, nm):
                matching: bool = getattr(state, nm) == val
                return matching
            return False

        for name, state in self._state_defs.items():
            if not isinstance(state, state_class):
                continue

            if attr_matches is None:
                return self._match_attr(name)

            has_attribute_matches = all(
                match_tester(nm, val, state) for nm, val in attr_matches.items()
            )
            if has_attribute_matches:
                return self._match_attr(name)
        return None

    def create_knowledge_event(
        self,
        *,
        name: str,
        rehearsal_time_to_complete: float = 0.0,
    ) -> Event:
        """Create an event and store it in knowledge.

        Useful for creating simple hold points in Tasks that can be succeeded by
        other processes.

        Example:
            >>> def task(self, actor):
            >>>     evt = actor.create_knowledge_event(name="hold")
            >>>     yield evt
            >>>     ... # do things
            ...
            >>> def other_task(self, actor):
            >>>     if condition:
            >>>         actor.succeed_knowledge_event(name="hold")

        Args:
            name (str): Name of the knowledge slot to store the event in.
            rehearsal_time_to_complete (float, optional): The event's expected
                time to complete. Defaults to 0.0.

        Returns:
            Event: The event to yield on
        """
        event = Event(rehearsal_time_to_complete=rehearsal_time_to_complete)
        # Rehearsals on this method won't clear the event, so save the user some trouble.
        overwrite = True if self._is_rehearsing else False
        self.set_knowledge(name, event, overwrite=overwrite)
        return event

    def succeed_knowledge_event(self, *, name: str, **kwargs: Any) -> None:
        """Succeed and clear an event stored in the actor's knowledge.

        See "create_knowledge_event" for usage example.

        Args:
            name (str): Event knowledge name.
            **kwargs (Any): Any payload to send to the event. Defaults to None
        """
        event = self.get_knowledge(name)
        if event is None:
            raise SimulationError(f"No knowledge named {name} to succeed")
        if not isinstance(event, Event):
            raise SimulationError(f"Knowledge {name} is not an Event.")
        self.clear_knowledge(name, "actor.succeed_knowledge_event")
        event.succeed(**kwargs)

    def get_remaining_waypoints(
        self, location_state: str
    ) -> list[GeodeticLocation] | list[CartesianLocation]:
        """Convenience method for interacting with LocationChangingStates.

        Primary use case is when restarting a Task that has a motion element to
        allow updating waypoint knowledge easily.

        Args:
            location_state (str): The name of the <LocationChangingState>

        Returns:
            list[Location]: List of waypoints yet to be reached
        """
        loc_state = self._state_defs[location_state]
        assert isinstance(loc_state, GeodeticLocationChangingState | CartesianLocationChangingState)
        wypts = loc_state._get_remaining_waypoints(self)
        return wypts

    def get_nucleus(self) -> "TaskNetworkNucleus":
        """Return the actor's nucleus.

        Returns:
            TaskNetworkNucleus: The nucleus on the actor.
        """
        if self._state_listener is None:
            raise SimulationError("Expected a nucleus, but none found.")
        return self._state_listener

    def record_state(self, state_name: str) -> None:
        """Record a state by its name.

        Useful for states that have attributes that aren't set
        via the descriptor, such as dictionaries or dataclasses.

        Args:
            state_name (str): The name of the state.
        """
        if state_name not in self.states:
            raise SimulationError(f"No state '{state_name}' to record.")
        v = getattr(self, state_name)
        self._state_defs[state_name]._do_record(self, v)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}: {self.name}"



"""Utilities for gathering all recorded simulation data."""

from dataclasses import asdict, is_dataclass
from typing import Any, cast

from upstage_des.actor import Actor
from upstage_des.base import UpstageBase
from upstage_des.data_types import CartesianLocation, GeodeticLocation
from upstage_des.states import (
    ActiveState,
    ActiveStatus,
    CartesianLocationChangingState,
    GeodeticLocationChangingState,
)

ACTUAL_LOCATION = GeodeticLocation | CartesianLocation
LOCATION_TYPES = ACTUAL_LOCATION | GeodeticLocationChangingState | CartesianLocationChangingState
STATIC_STATE = "Last Seen"

STATE_DATA_ROW = tuple[str, str, str, float, Any, str | None]
LOCATION_DATA_ROW = tuple[str, str, str, float, float, float, float, str | None]
COLUMN_NAMES = ["Entity Name", "Entity Type", "State Name", "Time"]
ACTIVATION_STATUS_COL = "Activation Status"


def _state_history_to_table(
    actor_name: str,
    actor_kind: str,
    state_name: str,
    is_active: bool,
    hist: list[tuple[float, Any]],
) -> list[STATE_DATA_ROW]:
    """Create a state history table from an actor.

    The final entry is a way to flag if a variable is becoming active or not.

    Args:
        actor_name (str): Actor name
        actor_kind (str): Actor kind
        state_name (str): State name
        is_active (bool): If the state is an active type
        hist (list[tuple[float, Any]]): History from _quantities or _state_histories

    Returns:
        list[STATE_DATA_ROW]: A long-form data table of state data.
    """
    data: list[STATE_DATA_ROW] = []
    active_value = "inactive" if is_active else None
    for time, value in hist:
        if isinstance(value, ActiveStatus):
            row = data.pop(-1)
            rows = [tuple(list(row[:-1]) + [value.name])]
            active_value = "active" if value.name == "activating" else "inactive"
        elif is_dataclass(value) and not isinstance(value, type):
            rows = [
                (actor_name, actor_kind, f"{state_name}.{k}", time, v, active_value)
                for k, v in asdict(value).items()
            ]
        elif isinstance(value, dict):
            rows = [
                (actor_name, actor_kind, f"{state_name}.{k}", time, v, active_value)
                for k, v in value.items()
            ]
        else:
            rows = [(actor_name, actor_kind, state_name, time, value, active_value)]
        data.extend(rows)
    return data


def _actor_state_data(
    actor: Actor,
    skip_locations: bool = True,
    save_static: bool = False,
) -> tuple[list[STATE_DATA_ROW], list[Any]]:
    """Gather actor recorded data.

    Args:
        actor (Actor): The actor.
        skip_locations (bool, optional): If location states should be ignored.
            Defaults to True.
        save_static (bool, optional): If non-recording states are saved.
            Defaults to False.

    Returns:
        list[STATE_INFO]: List of state information
        list[Any]: List of monitoring objects to ignore in a global search.
    """
    data: list[STATE_DATA_ROW] = []
    resources: list[Any] = []
    name, kind = actor.name, actor.__class__.__name__

    for state_name, state in actor._state_defs.items():
        if skip_locations and isinstance(state, LOCATION_TYPES):
            continue
        _value = actor.__dict__[state_name]
        is_active = isinstance(state, ActiveState)
        if state_name in actor._state_histories:
            data.extend(
                _state_history_to_table(
                    name, kind, state_name, is_active, actor._state_histories[state_name]
                )
            )
        elif hasattr(_value, "_quantities"):
            resources.append(_value)
            data.extend(_state_history_to_table(name, kind, state_name, False, _value._quantities))
        elif save_static:
            data.append((name, kind, state_name, 0.0, getattr(actor, state_name), STATIC_STATE))

    return data, resources


def _actor_location_data(actor: Actor) -> tuple[list[LOCATION_DATA_ROW], list[str]]:
    """Get actor location data, if it exists.

    The actor needs to have recording Location states:
        * CartesianLocation(ChangingState)
        * GeodeticLocation(ChangingState)

    Args:
        actor (Actor): The actor.

    Returns:
        list[LOCATION_DATA_ROW]: Time and XYZ/LLA data.
        list[str]: name of XYZ/LLA
    """
    data: list[LOCATION_DATA_ROW] = []
    is_xyz = True
    name, kind = actor.name, actor.__class__.__name__
    for state_name, state_data in actor._state_histories.items():
        _state = actor._state_defs[state_name]
        if not isinstance(_state, LOCATION_TYPES):
            continue
        value: ACTUAL_LOCATION | ActiveStatus
        is_active = isinstance(_state, ActiveState)
        active_value = "inactive" if is_active else None
        for time, value in state_data:
            if isinstance(value, ActiveStatus):
                _row = data.pop(-1)
                row = tuple(list(_row[:-1]) + [value.name])
                active_value = "active" if value.name == "activating" else "inactive"
            elif isinstance(value, GeodeticLocation):
                row = (name, kind, state_name, time, value.lat, value.lon, value.alt, active_value)
                is_xyz = False
            elif isinstance(value, CartesianLocation):
                row = (name, kind, state_name, time, value.x, value.y, value.z, active_value)
            data.append(cast(LOCATION_DATA_ROW, row))
    cols = ["X", "Y", "Z"] if is_xyz else ["Lat", "Lon", "Alt"]
    return data, cols


def create_table(
    skip_locations: bool = True, save_static: bool = False
) -> tuple[list[STATE_DATA_ROW], list[str]]:
    """Create a data table of everything UPSTAGE has recorded.

    This uses the current environment context.

    The data columns are:
        Time, Entity Name, Entity Type, State Name, State Value

    For SelfMonitoring<> resources that are not part of an actor, the name
    is pulled from the name entry to the resource. The Entity Type is the
    class name, and the State Name is "Resource State".

    Usage:

    >>> import pandas as pd
    >>> with UP.EnvironmentContext() as env:
    >>>     ...
    >>>     env.run()
    >>>     table, cols = create_table()
    >>>     df = pd.DataFrame(table, cols)

    Args:
        skip_locations (bool, optional): If location states should be ignored.
            Defaults to True.
        save_static (bool, optional): If non-recording states are saved.
            Defaults to False.

    Returns:
        list[STATE_DATA_ROW]: Data table
        list[str]]: Column names.
    """
    _base = UpstageBase()
    data: list[tuple[Any, ...]] = []
    seen_resources: list[Any] = []
    for actor in _base.get_actors():
        name = actor.name
        kind = actor.__class__.__name__
        _data, _resources = _actor_state_data(
            actor, skip_locations=skip_locations, save_static=save_static
        )
        seen_resources.extend(_resources)
        data.extend(_data)

    for monitoring in _base.get_monitored():
        if monitoring in seen_resources:
            continue
        name = f"{monitoring.name}"
        kind = f"{monitoring.__class__.__name__}"
        rows = [(name, kind, "Resource", t, value, None) for t, value in monitoring._quantities]
        data.extend(rows)

    colnames = COLUMN_NAMES + ["Value", ACTIVATION_STATUS_COL]
    return data, colnames


def create_location_table() -> tuple[list[LOCATION_DATA_ROW], list[str]]:
    """Create a data table of every location UPSTAGE has recorded.

    Assumes that all location types are the same.

    This uses the current environment context.

    Usage:

    >>> import pandas as pd
    >>> with UP.EnvironmentContext() as env:
    >>>     ...
    >>>     env.run()
    >>>     table, cols = create_location_table()
    >>>     df = pd.DataFrame(table, cols)

    Returns:
        list[LOCATION_DATA_ROW]: Data table
        list[str]]: Column names.
    """
    _base = UpstageBase()
    data: list[LOCATION_DATA_ROW] = []
    for actor in _base.get_actors():
        _data, cols = _actor_location_data(actor)
        data.extend(_data)
    return data, COLUMN_NAMES + cols + [ACTIVATION_STATUS_COL]



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""Conversions for common distance and time units."""

CONVERSIONS: dict[str, dict[str, float]] = {
    "m": {
        "km": 1 / 1000.0,
        "mi": 1 / 1000.0 * 0.62137119223,
        "nmi": 1 / 1000.0 * 0.539957,
        "m": 1.0,
    },
    "km": {"km": 1.0, "mi": 0.62137119223, "nmi": 0.539957, "m": 1000.0},
    "mi": {
        "km": 1 / 0.62137119223,
        "mi": 1.0,
        "nmi": 0.868976,
        "m": 1 / 0.62137119223 * 1000,
    },
    "nmi": {
        "km": 1 / 0.539957,
        "mi": 1 / 0.868976,
        "nmi": 1.0,
        "m": 1 / 0.539957 * 1000,
    },
    "s": {"hr": 1.0 / (60.0 * 60.0), "min": 1.0 / 60.0, "s": 1.0},
    "min": {"hr": 1.0 / 60.0, "min": 1.0, "s": 60.0},
    "hr": {"hr": 1.0, "min": 60.0, "s": (60.0 * 60.0)},
    "day": {"hr": 24.0, "week": 1 / 7.0, "min": 24 * 60.0, "s": 24 * 3600},
    "week": {"day": 7.0, "hr": 7.0 * 24, "min": 7.0 * 24 * 60.0, "s": 7.0 * 24 * 3600},
}

DISTANCE_UNITS = ["m", "km", "mi", "nmi", "ft"]
TIME_UNITS = ["s", "min", "hr", "day", "week"]
TIME_ALTERNATES = {
    "seconds": "s",
    "second": "s",
    "minute": "min",
    "minutes": "min",
    "hour": "hr",
    "hours": "hr",
    "days": "day",
    "weeks": "week",
}
STANDARD_TIMES = ["s", "min", "hr"]

# add feet to conversions
CONVERSIONS["ft"] = {"ft": 1.0, "mi": 1 / 5280.0}
for unit in DISTANCE_UNITS:
    if unit == "ft":
        continue
    CONVERSIONS["ft"][unit] = CONVERSIONS["mi"][unit] / 5280.0
    CONVERSIONS[unit]["ft"] = 1 / CONVERSIONS["ft"][unit]


def unit_convert(value: int | float, units_from: str, units_to: str) -> float:
    """Convert between units of distance and time.

    Units must be one of:
        distance: km, m, mi, nmi, ft

        time:

        * s, second, or seconds
        * min, minute, or minutes
        * hr, hour, or hours
        * day or days
        * week or weeks

    All units are lower-cased on input.

    Args:
        value (int | float): Value of "from" unit
        units_from (str): Unit to convert from
        units_to (str): Unit to convert to

    Raises:
        ValueError: _description_

    Returns:
        float: _description_
    """
    units_fr = units_from.lower()
    units_t = units_to.lower()
    units_fr = TIME_ALTERNATES.get(units_fr, units_fr)
    units_t = TIME_ALTERNATES.get(units_t, units_t)
    try:
        return value * CONVERSIONS[units_fr][units_t]
    except KeyError:
        raise ValueError(f"Cannot convert from {units_from} to {units_to}.")



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""Functions for finding intersections in geodetics."""

from upstage_des.math_utils import _vector_norm, _vector_subtract
from upstage_des.units import unit_convert

from .geo_types import LAT_LON_ALT, POSITION, POSITIONS, CrossingCondition
from .spherical import Spherical
from .wgs84 import WGS84


class IntersectionError(Exception):
    """An exception when intersection code fails."""

    ...


def _preprocess(
    start_lla: LAT_LON_ALT,
    finish_lla: LAT_LON_ALT,
    point_lla: LAT_LON_ALT,
    earth: Spherical | WGS84,
    dist_between: float,
) -> tuple[POSITIONS, POSITIONS, POSITION]:
    """Preprocess points for intersection calculations.

    Args:
        start_lla (LAT_LON_ALT): Starting point (degrees/meters)
        finish_lla (LAT_LON_ALT): Ending point (degrees/meters)
        point_lla (LAT_LON_ALT): Point to convert to ecef (degrees/meters)
        earth (Spherical | WGS84): Earth model
        dist_between (float): Distance to use for segment length (m)

    Returns:
        tuple[POSITIONS, POSITIONS, POSITION]: Preprocessed locations
    """
    start_alt = start_lla[2]
    finish_alt = finish_lla[2]
    dist = earth.distance(start_lla[:2], finish_lla[:2], units="m")
    points = int(dist / dist_between) + 1
    while points <= 2:
        dist_between = dist_between / 2.0
        if dist_between < 100:
            raise IntersectionError(
                f"Intersetion Segment {start_lla} -> {finish_lla} is too small!"
            )
        points = int(dist / dist_between) + 1
    ecef_point = earth.lla2ecef([point_lla])[0]
    assert len(ecef_point) == 3
    ecef_test, geo_test = earth.ecef_and_geo_linspace(
        start_lla[:2],
        finish_lla[:2],
        start_alt,
        finish_alt,
        points,
    )
    return ecef_test, geo_test, ecef_point


def find_crossing_points(
    start_lla: LAT_LON_ALT,
    finish_lla: LAT_LON_ALT,
    point_lla: LAT_LON_ALT,
    earth: Spherical | WGS84,
    radius: float,
    dist_between: float = 9260.0,
) -> list[CrossingCondition]:
    """Finds the points along a great circle path and a sphere.

    The output data provides booleans to state If the start or end are within
    range/visibility.

    Args:
        start_lla (LAT_LON_ALT): Starting point (degrees/meters)
        finish_lla (LAT_LON_ALT): Ending point (degrees/meters)
        point_lla (LAT_LON_ALT): Point of sensing (degrees/meters)
        earth (Spherical | WGS84): Earth model
        radius (float): Radius that the sensor can see (meters)
        dist_between (float, optional): Distance to use for segment length (meters).
            Defaults to 5 nmi (or 9260 meters).

    Returns:
        list[CrossingCondition]:
            A list of data describing the crossover points on the great circle path.
            It will start with: ["START_INSIDE" or "START_OUT", start LLA].
            Then there will be one or two: ["ENTER" or "EXIT", LLA, LLA] where the
            two LLA values are the OUT and IN points as described.
            It will end with: ["END_INSIDE" or "END_OUT", end LLA].
    """
    ecef_test, lla_test, ecef_point = _preprocess(
        start_lla,
        finish_lla,
        point_lla,
        earth,
        dist_between=dist_between,
    )

    last_out: int | None = None
    last_in: int | None = None
    cross_points: list[CrossingCondition] = []

    for i, test_loc in enumerate(ecef_test):
        diff = _vector_subtract(test_loc, ecef_point)
        dist = _vector_norm(diff)
        is_in = dist <= radius

        if i == 0:
            use = "START_INSIDE" if is_in else "START_OUT"
            cond = CrossingCondition(kind=use, begin=lla_test[i])
            cross_points.append(cond)
        elif is_in and last_in != i - 1:
            assert last_out is not None
            cond = CrossingCondition(
                kind="ENTER",
                begin=lla_test[last_out],
                end=lla_test[i],
            )
            cross_points.append(cond)
            assert last_out == i - 1
        elif not is_in and last_out != i - 1:
            assert last_in is not None
            cond = CrossingCondition(
                kind="EXIT",
                begin=lla_test[last_in],
                end=lla_test[i],
            )
            cross_points.append(cond)
            assert last_in == i - 1

        if is_in:
            last_in = i
        else:
            last_out = i

    use = "END_INSIDE" if is_in else "END_OUT"
    cond = CrossingCondition(kind=use, begin=lla_test[-1])
    cross_points.append(cond)

    return cross_points


def _split_down(
    begin: LAT_LON_ALT,
    end: LAT_LON_ALT,
    sphere_center: LAT_LON_ALT,
    radius: float,
    earth: Spherical | WGS84,
    distance_between: float,
    subdivide_levels: list[int] | None = None,
) -> CrossingCondition:
    """Find an intersection point from a sphere to a great circle.

    Default subdivision is [10, 20]

    We assume this is being called only on one clear crossing point, because it
    expects to find START_(IN|OUT) -> (IN_OUT | OUT_IN)  -> (END_(OUT|IN)) from the
    calls to find_crossing_points

    Args:
        begin (LAT_LON_ALT): Starting point (degrees/meters)
        end (LAT_LON_ALT): Ending point (degrees/meters)
        sphere_center (LAT_LON_ALT): Center of the sensing sphere (not earth, degrees/meters)
        radius (float): Sensor radius (meters)
        earth (Spherical | WGS84): Geodetic description
        distance_between (float): Splitting distance for search
        subdivide_levels (list[int], optional): Levels for searching smaller sections.
            Defaults to None.

    Returns:
        tuple[str, LAT_LON_ALT]: The intersection type, if any (degrees/meters)
    """
    subdivide_levels = [10, 20] if subdivide_levels is None else subdivide_levels

    for divide in subdivide_levels:
        distance_between = distance_between / divide
        split_data = find_crossing_points(
            begin,
            end,
            sphere_center,
            earth,
            radius,
            distance_between,
        )
        if len(split_data) != 3:
            raise IntersectionError(
                "A subdivide split shouldn't have 2 crossovers in intersections with a sphere"
            )
        if split_data[1].kind not in ["EXIT", "ENTER"]:
            raise IntersectionError("Subdividing an intersection check gave an invalid Direction")
        assert split_data[1].end is not None
        begin, end = split_data[1].begin, split_data[1].end

    cond = CrossingCondition(
        split_data[1].kind,
        begin=split_data[1].begin,
        end=split_data[1].end,
    )
    return cond


def get_intersection_locations(
    start: LAT_LON_ALT,
    finish: LAT_LON_ALT,
    sphere_center: LAT_LON_ALT,
    radius: float,
    radius_units: str,
    earth: WGS84 | Spherical,
    dist_between: float | None = None,
    subdivide_levels: list[int] | None = None,
) -> list[CrossingCondition]:
    """Get the locations and kinds of intersections of a path to a sphere.

    Args:
        start (LAT_LON_ALT): Starting point (degrees)
        finish (LAT_LON_ALT): Ending point (degrees)
        sphere_center (LAT_LON_ALT): Center of the sensing sphere (not earth, degrees/meters)
        radius (float): Sensor radius
        radius_units (str): Units of the sensor radius
        earth (Spherical | WGS84): Geodetic description
        dist_between (float, optional): Splitting distance for search.
            Defaults to 9260.0 meters (5 nmi)
        subdivide_levels (list[int], optional): Levels for searching smaller sections.
            Defaults to None.

    Returns:
        list[CrossingCondition]: The intersection type, if any
            It will start with:
                ["START_INSIDE" or "START_OUT"]
            Then there will be one or two:
                ["ENTER" or "EXIT"]
                where the two LLA values are the OUT and IN points as described
            It will end with:
                ["END_INSIDE" or "END_OUT"]
    """
    dist_between = 9260.0 if dist_between is None else dist_between
    radius = unit_convert(radius, radius_units, "m")
    subdivide_levels = [10, 20] if subdivide_levels is None else subdivide_levels
    rough_split_data = find_crossing_points(
        start,
        finish,
        sphere_center,
        earth,
        radius,
        dist_between,
    )

    # subdivide the splits for precision
    intersections: list[CrossingCondition] = []
    for condition in rough_split_data[1:-1]:
        assert condition.end is not None
        new_condition = _split_down(
            condition.begin,
            condition.end,
            sphere_center,
            radius,
            earth,
            dist_between,
            subdivide_levels=subdivide_levels,
        )
        if new_condition.kind != condition.kind:
            raise IntersectionError("Subdividing the intersection switched directions")
        use_lla = new_condition.begin if condition.kind == "EXIT" else new_condition.end
        assert use_lla is not None
        intersections.append(CrossingCondition(kind=condition.kind, begin=use_lla))

    # for START_IN and END_IN, append that data, too
    if rough_split_data[0].kind == "START_INSIDE":
        intersections = [rough_split_data[0]] + intersections
    if rough_split_data[-1].kind == "END_INSIDE":
        intersections = intersections + [rough_split_data[-1]]
    return intersections



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from math import degrees, pi, radians

import pytest

import upstage_des.api as UP
from upstage_des.motion.great_circle_calcs import (
    get_course_rad,
    get_dist_rad,
    get_great_circle_points,
    get_pos_from_points_and_distance,
)


def test_distance() -> None:
    """Test great circle distance calc."""
    with UP.EnvironmentContext():
        p1 = UP.GeodeticLocation(0, 180, 0).to_radians()
        p2 = UP.GeodeticLocation(1, 180, 0).to_radians()
        assert pytest.approx(get_dist_rad(p1, p2)) == pi / 180

        p3 = UP.GeodeticLocation(0, 180, 0).to_radians()
        p4 = UP.GeodeticLocation(0, 181, 0).to_radians()
        assert pytest.approx(get_dist_rad(p3, p4)) == pi / 180


def test_course() -> None:
    """Test great circle course calc."""
    with UP.EnvironmentContext():
        p1 = UP.GeodeticLocation(0, 180, 0).to_radians()
        p2 = UP.GeodeticLocation(1, 180, 0).to_radians()
        assert pytest.approx(get_course_rad(p1, p2)) == 2 * pi

        p3 = UP.GeodeticLocation(0, 180, 0).to_radians()
        p4 = UP.GeodeticLocation(0, 181, 0).to_radians()
        assert pytest.approx(get_course_rad(p3, p4)) == 3 * pi / 2.0


def test_position_from_point_distance() -> None:
    """Test position from point and distance calc."""
    with UP.EnvironmentContext():
        p1 = UP.GeodeticLocation(0, 180, 0).to_radians()
        p2 = UP.GeodeticLocation(1, 180, 0).to_radians()
        dist = 0.5 * (pi / 180)
        half_point = get_pos_from_points_and_distance(p1, p2, dist)
        assert pytest.approx(degrees(half_point[0])) == 0.5
        assert pytest.approx(degrees(half_point[1])) == -180


def test_great_circle_points() -> None:
    """Test calculation of points on creat circle path."""
    with UP.EnvironmentContext():
        p1 = UP.GeodeticLocation(0, 180, 0).to_radians()
        p2 = UP.GeodeticLocation(5, 180, 0).to_radians()
        p3 = UP.GeodeticLocation(3, 180, 0).to_radians()
        dist = pi / 180
        x = get_great_circle_points(p1, p2, p3, dist)
        assert x is not None
        points, distances = x
        assert len(points) == 2

        assert pytest.approx(points[0][0]) == radians(2)
        assert pytest.approx(points[0][1]) == radians(-180)
        assert pytest.approx(points[1][0]) == radians(4)
        assert pytest.approx(points[1][1]) == radians(-180)
        assert pytest.approx(distances[0]) == radians(2)
        assert pytest.approx(distances[1]) == radians(4)


def test_caching() -> None:
    with UP.EnvironmentContext():
        get_dist_rad.cache_clear()
        get_course_rad.cache_clear()
        get_pos_from_points_and_distance.cache_clear()
        get_great_circle_points.cache_clear()

        p1 = UP.GeodeticLocation(0, 180, 0).to_radians()
        p2 = UP.GeodeticLocation(1, 180, 0).to_radians()
        p3 = UP.GeodeticLocation(3, 180, 0).to_radians()
        dist = pi / 180

        get_dist_rad(p1, p2)
        get_dist_rad(p1, p2)
        cache_info = get_dist_rad.cache_info()
        assert cache_info.hits == 1
        assert cache_info.currsize == 1

        get_course_rad(p1, p2)
        get_course_rad(p1, p2)
        cache_info = get_course_rad.cache_info()
        assert cache_info.hits == 1
        assert cache_info.currsize == 1

        get_pos_from_points_and_distance(p1, p2, dist)
        get_pos_from_points_and_distance(p1, p2, dist)
        cache_info = get_pos_from_points_and_distance.cache_info()
        assert cache_info.hits == 1
        assert cache_info.currsize == 1

        get_great_circle_points(p1, p2, p3, dist)
        get_great_circle_points(p1, p2, p3, dist)
        cache_info = get_great_circle_points.cache_info()
        assert cache_info.hits == 1
        assert cache_info.currsize == 1



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

import pytest
import simpy as SIM

import upstage_des.api as UP
from upstage_des.type_help import SIMPY_GEN

from .flyer import Flyer, flyer_refuel_factory, mission_plan_net
from .mothership import Mothership, crew_factory, give_fuel_factory
from .mover import fly_end_factory


def build_sim() -> tuple[float, Mothership]:
    MEETING_POINT = UP.CartesianLocation(0, 0)
    TRAVERSE_TO_POINT = UP.CartesianLocation(200, 0)
    START_POINT = UP.CartesianLocation(200, 200)

    mothership = Mothership(
        name="Fuel Giver",
        fuel=1000.0,
        fuel_burn=-100.0,
        speed=200.0,
        location=START_POINT,
        fuel_ports_in_use=0,
        fuel_ports_max=4,
    )
    fly_net = fly_end_factory.make_network()
    give_fuel_net = give_fuel_factory.make_network()
    crew_net = crew_factory.make_network()
    for net in [fly_net, give_fuel_net, crew_net]:
        mothership.add_task_network(net)
        # assume that the first entered task name is the one we want
        mothership.start_network_loop(
            net.name,
            list(net.task_classes.keys())[0],
        )

    # give some knowledge
    mothership.set_knowledge(
        "destinations",
        [MEETING_POINT, TRAVERSE_TO_POINT, START_POINT],
    )

    # nucleus
    mothership_nucleus = UP.TaskNetworkNucleus(actor=mothership)
    mothership_nucleus.add_network(
        fly_net,
        [
            "speed",
        ],
    )
    mothership_nucleus.add_network(give_fuel_net, ["fuel_ports_in_use"])

    total_dist = (
        (START_POINT - TRAVERSE_TO_POINT)
        + (TRAVERSE_TO_POINT - MEETING_POINT)
        + (START_POINT - MEETING_POINT)
    )
    return total_dist, mothership


def speed_change(
    env: SIM.Environment,
    vehicle: Mothership,
    new_speed: float,
    time: float,
) -> SIMPY_GEN:
    yield env.timeout(time)
    vehicle.speed = new_speed


def add_draw(
    env: SIM.Environment,
    vehicle: Mothership,
    amount: float,
    time_to: float,
    time_on: float,
) -> SIMPY_GEN:
    class Dummy(UP.Actor):
        messages = UP.ResourceState[SIM.Store](default=SIM.Store)

    d = Dummy(name="a_vehicle")
    yield env.timeout(time_to)
    yield vehicle.messages.put((d, amount))
    yield env.timeout(time_on)
    yield vehicle.messages.put((d, 0))


def test_nothing_added() -> None:
    with UP.EnvironmentContext() as env:
        total_dist, mothership = build_sim()
        t = total_dist / 200
        env.run()
        assert pytest.approx(env.now) == t
        assert pytest.approx(mothership.fuel) == 1000 - 100 * t


def test_fueling() -> None:
    with UP.EnvironmentContext() as env:
        _, mothership = build_sim()

        env.process(speed_change(env, mothership, 150, 1.3))
        env.process(add_draw(env, mothership, -100, 0.5, 1.5))
        env.process(add_draw(env, mothership, -200, 1.0, 2))

        env.run()

        # final time must be larger (~3.4 for no speed change)
        assert pytest.approx(4.1189514164974605) == env.now

        # final fuel level
        assert pytest.approx(63.104858350253835) == mothership.fuel


# Fake mothership
class Dummy(UP.Actor):
    messages = UP.ResourceState[SIM.Store](default=SIM.Store)


def figher_build() -> tuple[Flyer, Dummy]:
    MEETING_POINT = UP.CartesianLocation(0, 0)
    TRAVERSE_TO_POINT = UP.CartesianLocation(200, 0)
    START_POINT = UP.CartesianLocation(-200, -200)

    flyer = Flyer(
        name="Flyer Thing",
        fuel=1000.0,
        fuel_capacity=1500.0,
        fuel_burn=-100.0,
        fuel_draw=500,
        speed=300.0,
        location=START_POINT,
        debug_log=True,
    )

    fly_net = fly_end_factory.make_network()
    get_fuel_net = flyer_refuel_factory.make_network()
    plan_net = mission_plan_net.make_network()
    for net in [fly_net, get_fuel_net, plan_net]:
        flyer.add_task_network(net)
        # assume that the first entered task name is the one we want
        flyer.start_network_loop(
            net.name,
            list(net.task_classes.keys())[0],
        )

    # give some knowledge
    flyer.set_knowledge(
        "destinations",
        [MEETING_POINT, TRAVERSE_TO_POINT, START_POINT],
    )

    d = Dummy(name="the_mothership")
    # env.process(dummy_mothership_proc(env, d))

    flyer.set_knowledge("mothership", d)
    flyer.set_knowledge("refuel_speed", 150)
    flyer.set_knowledge("meetup", MEETING_POINT)

    # nucleus
    flyer_nucleus = UP.TaskNetworkNucleus(actor=flyer)
    flyer_nucleus.add_network(
        fly_net,
        [
            "speed",
        ],
    )
    flyer_nucleus.add_network(get_fuel_net, ["approach"])
    return flyer, d


def dummy_mothership_proc(mothership: Dummy) -> SIMPY_GEN:
    msg = yield mothership.messages.get()
    sendback = msg[0]
    yield sendback.messages.put("GO")


def test_flyer_nothing() -> None:
    with UP.EnvironmentContext() as env:
        flyer, _ = figher_build()
        env.run()
        t = 5.257566344915117
        assert pytest.approx(env.now) == t
        assert pytest.approx(flyer.fuel) == 1000 - 100 * t


def test_flyer_refuel() -> None:
    with UP.EnvironmentContext() as env:
        flyer, mothership = figher_build()
        env.process(dummy_mothership_proc(mothership))
        env.run()
        t = 5.257566344915117
        assert pytest.approx(env.now) == t
        assert pytest.approx(flyer.fuel) == 1068.5242696666946


def test_full_fuelinging() -> None:
    with UP.EnvironmentContext() as env:
        MEETING_POINT = UP.CartesianLocation(0, 0)
        TRAVERSE_TO_POINT = UP.CartesianLocation(200, 0)
        START_POINT = UP.CartesianLocation(200, 200)

        mothership = Mothership(
            name="Fuel Delivery",
            fuel=1000.0,
            fuel_burn=-100.0,
            speed=200.0,
            location=START_POINT,
            fuel_ports_in_use=0,
            fuel_ports_max=4,
        )
        fly_net = fly_end_factory.make_network()
        give_fuel_net = give_fuel_factory.make_network()
        crew_net = crew_factory.make_network()
        for net in [fly_net, give_fuel_net, crew_net]:
            mothership.add_task_network(net)
            # assume that the first entered task name is the one we want
            mothership.start_network_loop(
                net.name,
                list(net.task_classes.keys())[0],
            )

        # give some knowledge
        mothership.set_knowledge(
            "destinations",
            [MEETING_POINT, TRAVERSE_TO_POINT, START_POINT],
        )

        # nucleus
        mothership_nucleus = UP.TaskNetworkNucleus(actor=mothership)
        mothership_nucleus.add_network(
            fly_net,
            [
                "speed",
            ],
        )
        mothership_nucleus.add_network(give_fuel_net, ["fuel_ports_in_use"])

        MEETING_POINT = UP.CartesianLocation(0, 0)
        TRAVERSE_TO_POINT = UP.CartesianLocation(200, 0)
        START_POINT = UP.CartesianLocation(-200, -200)

        flyer = Flyer(
            name="Flying Thing",
            fuel=1000.0,
            fuel_capacity=1500.0,
            fuel_burn=-100.0,
            fuel_draw=500,
            speed=300.0,
            location=START_POINT,
            debug_log=True,
        )

        fly_net = fly_end_factory.make_network()
        get_fuel_net = flyer_refuel_factory.make_network()
        plan_net = mission_plan_net.make_network()
        for net in [fly_net, get_fuel_net, plan_net]:
            flyer.add_task_network(net)
            # assume that the first entered task name is the one we want
            flyer.start_network_loop(
                net.name,
                list(net.task_classes.keys())[0],
            )

        # give some knowledge
        flyer.set_knowledge(
            "destinations",
            [MEETING_POINT, TRAVERSE_TO_POINT, START_POINT],
        )

        flyer.set_knowledge("mothership", mothership)
        # This speed doesn't match the mothership, but
        # the test numbers are tuned for it
        flyer.set_knowledge("refuel_speed", 150)
        flyer.set_knowledge("meetup", MEETING_POINT)

        # nucleus
        flyer_nucleus = UP.TaskNetworkNucleus(actor=flyer)
        flyer_nucleus.add_network(
            fly_net,
            [
                "speed",
            ],
        )
        flyer_nucleus.add_network(get_fuel_net, ["approach"])

        env.run()

        # The total fuel to start is: 2000
        flyer_flight_t = 5.257566344915117
        flyer_burn = 100 * flyer_flight_t
        mothership_flight_t = 682.842712474619 / 200
        mothership_burn = 100 * mothership_flight_t
        total_burn = mothership_burn + flyer_burn
        fuel_left = mothership.fuel + flyer.fuel
        assert pytest.approx(fuel_left) == 2000 - total_burn



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""Motion features for UPSTAGE."""

from .motion import MotionAndDetectionError, SensorMotionManager
from .stepped_motion import SteppedMotionManager

__all__ = ["MotionAndDetectionError", "SensorMotionManager", "SteppedMotionManager"]
